from __future__ import annotations

import logging
from typing import Optional, List, TYPE_CHECKING

from vectice.api.json import (
    JobType,
    ModelType,
    DatasetInput,
    ModelInput,
    JobInput,
    CodeVersionInput,
    Page,
    StageInput,
    StageStatus,
    StageOrigin,
)
from .code_version import CodeVersion
from .connection import Connection
from .data_resource import DataResource
from .dataset import Dataset
from .git_version import GitVersion
from .integration import AbstractIntegration
from .job import Job
from .model import Model
from .stage import Stage

if TYPE_CHECKING:
    from vectice import Reference
    from vectice.models import Workspace


class Project:
    """
    A project contains resources relative to a specific project.

    A project can contain :

    + documentation describing the project (goals, status, ...)
    + several datasets,
    + several models
    + several jobs

    A project can contains several datasets. Each datasets contains its own history.
    This is the same for models and jobs.

    """

    def __init__(
        self,
        id: int,
        workspace: Workspace,
        name: str,
        description: Optional[str] = None,
    ):
        """

        :param id: Project identifier
        :param workspace: the workspace reference this project belong to
        :param name: Name of the project
        :param description: Quick description of the project
        """
        self._id = id
        self._workspace = workspace
        self._name = name
        self._description = description
        self._client = workspace._client
        self._integration_client: Optional[AbstractIntegration] = workspace._integration_client
        self._logger = logging.getLogger(self.__class__.__name__)

    def __repr__(self):
        return f"Project(name={self.name}, id={self.id}, description={self.description}, workspace={self.workspace})"

    def update(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None:
        """
        Updates the information of the project.

        :param name: The name of the project.
        :param description: the description of the project

        :return: None
        """
        pass

    @property
    def id(self) -> int:
        """
        Project identifier.
        :return: int
        """
        return self._id

    @property
    def workspace(self) -> Workspace:
        """
        The workspace object this project belong to.
        :return: Workspace
        """
        return self._workspace

    @property
    def name(self) -> str:
        """
        Name of the project.
        :return: str
        """
        return self._name

    @property
    def description(self) -> Optional[str]:
        """
        Quick description of the project.
        :return: Optional[str]
        """
        return self._description

    def delete(self) -> None:
        """
        Indicates if the current project is deleted

        :return: None
        """
        pass

    def list_jobs(
        self,
        search: Optional[str] = None,
        page_index: int = 1,
        page_size: int = 20,
    ) -> List[Job]:
        """
        Lists the jobs in this project

        The search parm helps filter the jobs depending on the name given.

        :param search: The name to search
        :param page_index: The page index
        :param page_size: The page size

        :return: List of Jobs
        """
        outputs = self._client.list_jobs(self.id, None, search, page_index, page_size)
        return [Job(item.name, item.id, self, item.description, item.type) for item in outputs.list]

    def delete_job(self, job: Reference) -> None:
        """
        Deletes the specified job in the current project.

        :param job: The job name or id

        :return: None
        """
        item = self._client.get_job(job, self.id)
        self._logger.info(f"Job with id: {item.id} successfully deleted.")
        self._client.delete_job(job, self.id)

    def get_job(self, job: Reference) -> Job:
        """
        Gets a job.

        :param job: The job name or id

        :return: A Job or None if not found
        """
        item = self._client.get_job(job, self.id)
        self._logger.info(f"Job with id: {item.id} successfully retrieved.")
        return Job(item.name, item.id, self, item.description, item.type)

    def create_job(
        self,
        name: str,
        description: Optional[str] = None,
        type: JobType = JobType.OTHER,
    ) -> Job:
        """
        Create a new job in the project with the given attributes.

        The name is required and must be unique in the project.

        The default job type is `OTHER`. check the :class:`JobType` for other possible values.

        :param name: name of the job
        :param description: quick description of the job
        :param type: the type of the job

        :return: the newly created Job
        """
        data = JobInput(name, description, type)
        output = self._client.create_job(data, self.id)
        self._logger.info(f"Job with id: {output.id} successfully created.")
        return Job(output.name, output.id, self, output.description, output.type)

    def update_job(
        self,
        job: Reference,
        name: Optional[str] = None,
        description: Optional[str] = None,
        type: Optional[JobType] = None,
    ) -> Job:
        """
        Update the specified job with the new attribute values.

        :param job: The name or identifier of the job to update
        :param name: The name of the job
        :param description: The description of the job
        :param type: The job type

        :return: Job
        """
        data = self._client.get_job(job, self.id)
        if name is not None:
            data["name"] = name
        if description is not None:
            data["description"] = description
        if type is not None:
            data["type"] = type
        output = self._client.update_job(data, self.id)
        self._logger.info(f"Job with id: {output.id} successfully updated.")
        return Job(output.name, output.id, self, output.description, output.type)

    def list_datasets(self, search: Optional[str] = None, page_index: int = 1, page_size: int = 20) -> List[Dataset]:
        """
        Lists the datasets of this project.

        The search param filter the datasets with the specified name.

        :param search: The dataset to search
        :param page_index:
        :param page_size:

        :return: List of Datasets
        """
        outputs = self._client.list_datasets(self.id, None, search, page_index, page_size)
        return [
            Dataset(
                item.name,
                item.id,
                self,
                item.description,
                Connection(
                    item.connection.id,
                    item.connection.name,
                    self.workspace,
                    item.connection.type,
                    item.connection.parameters,
                    item.connection.description,
                )
                if item.connection is not None
                else None,
            )
            for item in outputs.list
        ]

    def get_dataset(self, dataset: Reference) -> Dataset:
        """
        Gets a dataset.

        :param dataset: dataset name or id

        :return: A Dataset or None if not found
        """
        dataset_output = self._client.get_dataset(dataset, self.id, self.workspace.id)
        if dataset_output.connection_id is not None:
            connection_output = self._client.get_connection(dataset_output.connection_id, self.workspace.id)
            connection = Connection(
                id=connection_output.id,
                name=connection_output.name,
                workspace=self.workspace,
                type=connection_output.type,
                parameters=connection_output.parameters,
                description=connection_output.description,
            )
        else:
            connection = None
        self._logger.info(f"Dataset with id: {dataset_output.id} successfully retrieved.")
        return Dataset(
            dataset_output.name,
            dataset_output.id,
            project=self,
            description=dataset_output.description,
            connection=connection,
        )

    def create_dataset(
        self,
        name: str,
        pattern: str = "*",
        description: Optional[str] = None,
        connection: Optional[Reference] = None,
        resources: Optional[List[str]] = None,
    ) -> Dataset:
        """
        Create a new dataset in a specific project.

        :param name: The name of the dataset
        :param pattern: The dataset pattern
        :param description: The description of the dataset
        :param connection: The connection name or id the dataset belong to
        :param resources: The file/s of the dataset version

        :return: Dataset
        """
        data = DatasetInput(
            name=name,
            pattern=pattern,
            description=description,
            connectionId=connection if isinstance(connection, int) else None,
            connectionName=connection if isinstance(connection, str) else None,
        )
        if resources is not None and len(resources) > 0:
            data["dataResources"] = [DataResource.create_resource(item) for item in resources]
        dataset_output = self._client.create_dataset(data, self.id)
        if connection is not None:
            connection_output = self._client.get_connection(connection, self.workspace.id)
            connection_object = Connection(
                id=connection_output.id,
                name=connection_output.name,
                workspace=self.workspace,
                type=connection_output.type,
                parameters=connection_output.parameters,
                description=connection_output.description,
            )
        else:
            connection_object = None
        self._logger.info(f"Dataset with id: {dataset_output.id} successfully created.")
        return Dataset(
            dataset_output.name,
            dataset_output.id,
            project=self,
            description=dataset_output.description,
            connection=connection_object,
            resources=dataset_output.resources,
        )

    def update_dataset(
        self,
        dataset: Reference,
        name: Optional[str] = None,
        description: Optional[str] = None,
        connection: Optional[Reference] = None,
    ) -> Dataset:
        """
        Updates a dataset. The connection is associated with the resources, e.g gcs connection for a gcs resource.

        :param dataset: dataset name or id
        :param name: The name of the dataset version e.g Changes 'Version 1' to 'name'
        :param description: The description of the dataset version
        :param connection: The connection name or id

        :return: Dataset
        """
        dataset_output = self._client.get_dataset(dataset, self.id, self.workspace.id)
        if name is not None:
            dataset_output["name"] = name
        if description is not None:
            dataset_output["description"] = description
        if connection is not None:
            connection_output = self._client.get_connection(connection, self.workspace.id)
            dataset_output["connectionId"] = connection_output.id
            connection_object = Connection(
                id=connection_output.id,
                name=connection_output.name,
                workspace=self.workspace,
                type=connection_output.type,
                parameters=connection_output.parameters,
                description=connection_output.description,
            )
        else:
            connection_object = None
        dataset_output = self._client.update_dataset(dataset_output, dataset, self.id, self.workspace.id)
        self._logger.info(f"Dataset with id: {dataset_output.id} successfully updated.")
        return Dataset(
            dataset_output.name,
            dataset_output.id,
            project=self,
            description=dataset_output.description,
            connection=connection_object,
        )

    def delete_dataset(self, dataset: Reference) -> None:
        """
        Deletes the dataset with the specified reference.

        :param dataset: The dataset name or id

        :return: None
        """
        dataset_id = self._client.get_dataset(dataset, self.id, self.workspace.id).id
        self._logger.info(f"Dataset with id: {dataset_id} successfully deleted.")
        self._client.delete_dataset(dataset, self.id, self.workspace.id)

    def create_model(
        self,
        name: str,
        description: Optional[str] = None,
        type: Optional[ModelType] = ModelType.OTHER,
    ) -> Model:
        """
        Creates a new model in the specified project.
        :param name: The name of the model
        :param description: The description of the model
        :param type: The model type e.g

        :return: A Model
        """
        data = ModelInput(name=name, description=description, type=type)
        model_output = self._client.create_model(data, self.id)
        self._logger.info(f"Model with id: {model_output.id} successfully created.")
        return Model(model_output.name, model_output.id, self, model_output.description, ModelType[model_output.type])

    def update_model(
        self,
        model: Reference,
        name: Optional[str] = None,
        description: Optional[str] = None,
        type: Optional[ModelType] = ModelType.OTHER,
    ) -> Model:
        """
        Updates a model with the new attribute values.

        :param model: the name or identifier of the model to update
        :param name: The name of the model
        :param description: The description of the model
        :param type: The model type

        :return: Model
        """
        data = ModelInput(name=name, description=description, type=type)
        model_output = self._client.update_model(data, model, self.id)
        self._logger.info(f"Model with id: {model_output.id} successfully updated.")
        return Model(model_output.name, model_output.id, self, model_output.description, ModelType[model_output.type])

    def list_models(
        self, search: Optional[str] = None, page_index: int = Page.index, page_size: int = Page.size
    ) -> List[Model]:
        """
        Lists the models of the project

        The search param helps find the models with the specified name.

        :param search: The model to search
        :param page_index: The page index
        :param page_size:The page size

        :return: List of Models
        """
        outputs = self._client.list_models(self.id, None, search, page_index, page_size)
        return [Model(item.name, item.id, self, item.description, ModelType[item.type]) for item in outputs.list]

    def get_model(self, model: Reference) -> Model:
        """
        Gets a model from the specified project.

        :param model: The model name or id
        :return: A Model or None if not found
        """
        model_output = self._client.get_model(model, self.id)
        self._logger.info(f"Model with id: {model_output.id} successfully retrieved.")
        return Model(model_output.name, model_output.id, self, model_output.description, ModelType[model_output.type])

    def delete_model(self, model: Reference) -> None:
        """
        Deletes the model with the specified reference.

        :param model: The model name or id
        :return: None
        """
        model_id = self._client.get_model(model, self.id).id
        self._logger.info(f"Model with id: {model_id} successfully deleted.")
        self._client.delete_model(model, self.id)

    def create_code_version(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        uri: Optional[str] = None,
        is_starred: Optional[bool] = None,
        git_version: Optional[GitVersion] = None,
    ) -> CodeVersion:
        """
        Gets a specific version of this code.

        :param name: The version name or id
        :param description: A quick description of the code version
        :param uri: The uri of the code
        :param is_starred: If the code version is starred
        :param git_version: The git version of the code version

        :return: A CodeVersion or None if the version does not exist
        """
        logging.getLogger("Client").propagate = False
        code_version_input = CodeVersionInput(
            name=name,
            description=description,
            uri=uri,
            is_starred=is_starred,
            gitVersion=git_version,
        )
        code_version_output = self._client.create_code_version(code_version_input, self.id)
        code_name = self._client.get_code(code_version_output.code_id).name
        logging.getLogger("Client").propagate = True
        self._logger.info(f"CodeVersion with id: {code_version_output.id} successfully created in Code {code_name}.")
        return CodeVersion(
            project=self,
            id=code_version_output.id,
            name=code_version_output.name,
            description=code_version_output.description,
            version_number=code_version_output.version_number,
            uri=code_version_output.uri,
            is_starred=code_version_output.is_starred,
            git_version=GitVersion(
                repositoryName=code_version_output.git_version.repository_name,
                branchName=code_version_output.git_version.branch_name,
                isDirty=code_version_output.git_version.is_dirty,
                commitHash=code_version_output.git_version.commit_hash,
                commitComment=code_version_output.git_version.commit_comment,
                entrypoint=code_version_output.git_version.entrypoint,
                commitAuthorName=code_version_output.git_version.commit_author_name,
                commitAuthorEmail=code_version_output.git_version.commit_author_email,
                uri=code_version_output.git_version.uri,
            )
            if code_version_output.git_version is not None
            else None,
            version=code_version_output.version,
            code_id=code_version_output.code_id,
        )

    def get_code_version(
        self,
        version: Reference,
    ) -> CodeVersion:
        """
        Gets a specific version of this code.

        :param version: The version name or id

        :return: A CodeVersion or None if the version does not exist
        """
        logging.getLogger("Client").propagate = False
        code_version_output = self._client.get_code_version(version, self.id)
        code_name = self._client.get_code(code_version_output.code_id, self.id, self.workspace.id).name
        logging.getLogger("Client").propagate = True
        self._logger.info(
            f"CodeVersion with id: {code_version_output.id} successfully retrieved from Code {code_name}."
        )
        return CodeVersion(
            project=self,
            id=code_version_output.id,
            name=code_version_output.name,
            version_number=code_version_output.version_number,
            description=code_version_output.description,
            attachments="",
            is_starred=code_version_output.is_starred,
            version=code_version_output.version,
            git_version=GitVersion(
                repositoryName=code_version_output.git_version.repository_name,
                branchName=code_version_output.git_version.branch_name,
                isDirty=code_version_output.git_version.is_dirty,
                commitHash=code_version_output.git_version.commit_hash,
                commitComment=code_version_output.git_version.commit_comment,
                entrypoint=code_version_output.git_version.entrypoint,
                commitAuthorName=code_version_output.git_version.commit_author_name,
                commitAuthorEmail=code_version_output.git_version.commit_author_email,
                uri=code_version_output.git_version.uri,
            )
            if code_version_output.git_version is not None
            else None,
        )

    def list_code_versions(
        self, search: Optional[str] = None, page_index: int = 1, page_size: int = 20
    ) -> List[CodeVersion]:
        """
        Lists the code versions.

        Returns a List of the code Versions associated with the project.
        The CodeVersion can be used in runs or experiments.

        :param search: The name that contains the code versions
        :param page_index: The page index
        :param page_size: The page size

        :return: List of CodeVersions
        """

        code_version_output_list = self._client.list_code_versions(self.id, None, search, page_index, page_size)
        return [
            CodeVersion(
                project=self,
                id=code_version_output.id,
                name=code_version_output.name,
                version_number=code_version_output.version_number,
                description=code_version_output.description,
                is_starred=code_version_output.is_starred,
                git_version_id=code_version_output.git_version_id,
                version=code_version_output.version,
            )
            for code_version_output in code_version_output_list
        ]

    def update_code_version(
        self,
        version: Reference,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_starred: Optional[bool] = None,
        uri: Optional[str] = None,
        git_version: Optional[GitVersion] = None,
    ) -> CodeVersion:
        """
        Updates a code version.

        :param version: The code version name or id
        :param name: The name of the code version e.g Changes 'Version 1' to 'name'
        :param description: The description of the code version
        :param is_starred: Whether the dataset version is starred
        :param uri: The uri of the code version
        :param git_version: The GitVersion of the code version
        :return: The updated code version
        """
        logging.getLogger("Client").propagate = False
        code_version_object = self._client.get_code_version(version, self.id)
        if code_version_object is None:
            raise ValueError("The code version was not found. Please check the reference.")
        code_version_input = CodeVersionInput(
            name=name, description=description, isStarred=is_starred, uri=uri, GitVersion=git_version
        )
        code_version_output = self._client.update_code_version(code_version_input, code_version_object.id, self.id)
        code_name = self._client.get_code(code_version_output.code_id).name
        logging.getLogger("Client").propagate = True
        self._logger.info(f"CodeVersion with id: {code_version_output.id} successfully updated in Code {code_name}.")
        return CodeVersion(
            project=self,
            id=code_version_output.id,
            name=code_version_output.name,
            version_number=code_version_output.version_number,
            description=code_version_output.description,
            is_starred=code_version_output.is_starred,
            version=code_version_output.version_number,
            git_version_id=code_version_output.git_version_id,
        )

    def delete_code_version(
        self,
        version: Reference,
    ) -> None:
        """
        Deletes a specific code version.

        :param version: The code version name or id
        :return: None
        """
        self._client.delete_code_version(version, self.id)

    def create_stage(self, name: Optional[str] = None, status: Optional[StageStatus] = StageStatus.NotStarted) -> Stage:
        """
        Add a stage to the project. A documentation page is created with the title provided. The status can be 'Draft/InProgress/Completed'.

        :param name: The name of the stages documentation page.
        :param status: The status of the stage.

        :return: A Stage object
        """
        data = StageInput(name=name, status=status)
        stage_output = self._client.create_stage(data, self.id)
        self._logger.info(f"Stage '{stage_output.name}' created.")
        return Stage(
            stage_output.id,
            self,
            stage_output.name,
            StageStatus["InProgress" if stage_output.status == "Draft" else stage_output.status],
            StageOrigin[stage_output.origin],
        )

    def update_stage(
        self, stage: Reference, name: Optional[str] = None, status: Optional[StageStatus] = StageStatus.NotStarted
    ) -> Stage:
        """
        Update a stage of the project. A documentation page is updated with the title and status provided. The status can be 'Draft/InProgress/Completed'.

        :param stage: The name or id of the stage.
        :param name: The name of the stages documentation page.
        :param status: The status of the stage.

        :return: A Stage object
        """
        data = StageInput(name=name, status=status)
        stage_output = self._client.update_stage(stage, data, self.id)
        self._logger.info(f"Stage '{stage_output.name}' updated.")
        return Stage(
            stage_output.id,
            self,
            stage_output.name,
            "InProgress" if stage_output.status == "Draft" else StageStatus[stage_output.status],
            StageOrigin[stage_output.origin],
        )

    def get_stage(self, stage: Reference) -> Stage:
        """
        Get a stage of the project. A stage will be returned.

        :param stage: The name or id of the stage.

        :return: Stage
        """
        stage_output = self._client.get_stage(stage, self.id)
        self._logger.info(f"Stage '{stage_output.name}' retrieved.")
        return Stage(
            stage_output.id,
            self,
            stage_output.name,
            StageStatus["InProgress" if stage_output.status == "Draft" else stage_output.status],
            StageOrigin[stage_output.origin],
        )

    def list_stages(
        self, search: Optional[str] = None, page_index: int = Page.index, page_size: int = Page.size
    ) -> List[Stage]:
        """
        List all the stages of the project. A search term can be used.

        :param search: The search term to be used.

        :return: A list of stages
        """
        outputs = self._client.list_stages(self.id, None, search, page_index, page_size)
        self._logger.info("Stages successfully retrieved.")
        return [
            Stage(
                item.id,
                self,
                item.name,
                StageStatus["InProgress" if item.status == "Draft" else item.status],
                StageOrigin[item.origin],
            )
            for item in outputs
        ]

    def delete_stage(self, stage: Reference) -> None:
        """
        Delete a stage of the project. A documentation page related to the stage will be deleted aswell.

        :param stage: The name or id of the stage.

        :return: None
        """
        self._client.delete_stage(stage, self.id)
        self._logger.info(f"Stage '{stage}' successfully deleted.")



from . import models
from . import api
from .api import Reference
from .integrations import Integrations

from .vectice import Vectice
from vectice.experiment import Experiment
from .__version__ import __version__


_AUTOLOGGING = []

try:
    import mlflow.sklearn as sklearn
    import mlflow.tracking.fluent.autolog as autolog

    _AUTOLOGGING = ["autolog", "sklearn"]

except ImportError:
    pass

version = __version__

__all__ = ["Vectice", "api", "Integrations", "models", "Experiment", "version"] + _AUTOLOGGING



from typing import Optional

from vectice.api import Reference
from vectice.api.json import ArtifactType, VersionStrategy


class ArtifactReference:
    def __init__(
        self,
        code: Optional[Reference] = None,
        dataset: Optional[Reference] = None,
        model: Optional[Reference] = None,
        version_number: Optional[int] = None,
        version_id: Optional[int] = None,
        version_name: Optional[str] = None,
        version_strategy: VersionStrategy = VersionStrategy.MANUAL,
        description: Optional[str] = None,
    ):
        self.dataset = dataset
        self.model = model
        self.code = code
        self.version_number = version_number
        self.version_name = version_name
        self.version_id = version_id
        self.version_strategy = version_strategy
        self.description = description

    def __repr__(self):
        return (
            "ArtifactReference("
            + f"code={self.code}, "
            + f"dataset={self.dataset}, "
            + f"model={self.model}, "
            + f"version_number={self.version_number}, "
            + f"version_id={self.version_id}, "
            + f"version_name={self.version_name}, "
            + f"version_strategy={self.version_strategy}, "
            + f"description={self.description}, "
            + ")"
        )

    def __eq__(self, other):
        return (
            isinstance(other, ArtifactReference)
            and self.code == other.code
            and self.dataset == other.dataset
            and self.model == other.model
            and self.version_number == other.version_number
            and self.version_id == other.version_id
            and self.version_name == other.version_name
            and self.version_strategy == other.version_strategy
            and self.description == other.description
        )

    @property
    def artifact_type(self) -> ArtifactType:
        if self.dataset is not None:
            return ArtifactType.DATASET
        if self.model is not None:
            return ArtifactType.MODEL
        if self.code is not None:
            return ArtifactType.CODE
        raise RuntimeError("empty artifact")



from vectice.api._auth import Auth
from vectice.api.http_error_handlers import HttpErrorHandler
from vectice.api.json.compatibility import CompatibilityOutput


class CompatibilityApi:
    def __init__(self, auth: Auth):
        self._auth = auth
        self._httpErrorhandler = HttpErrorHandler()

    def check_version(self) -> CompatibilityOutput:
        response = self._auth._get("/metadata/compatibility")
        if "message" not in response.keys():
            return CompatibilityOutput(message="", status="OK")
        else:
            return CompatibilityOutput(**response)



from __future__ import annotations

import logging
import urllib
from typing import Optional, TYPE_CHECKING, List
from urllib.parse import urlencode

from ._auth import Auth
from ._http import HttpError
from .http_error_handlers import HttpErrorHandler
from .json import (
    ModelVersionOutput,
    ModelVersionInput,
    Page,
    PagedResponse,
    PropertyInput,
    PropertyOutput,
    MetricInput,
    MetricOutput,
)

from .model import ModelApi
from .reference import InvalidReferenceError, MissingReferenceError

if TYPE_CHECKING:
    from vectice import Reference


class ModelVersionApi:
    def __init__(self, auth: Auth):
        self._auth = auth
        self._httpErrorhandler = HttpErrorHandler()
        self._logger = logging.getLogger(self.__class__.__name__)

    def get_model_version(
        self,
        version: Reference,
        model: Optional[Reference],
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> ModelVersionOutput:
        if not isinstance(version, int) and not isinstance(version, str):
            raise ValueError("The model version reference is invalid. Please check the entered value.")

        if isinstance(version, int):
            url = f"/metadata/modelversion/{version}"
        elif isinstance(version, str):
            if model is None:
                raise MissingReferenceError("model version", "model")
            parent_model = ModelApi(self._auth).get_model(model, project, workspace)
            url = f"/metadata/project/{parent_model.project.id}/model/{parent_model.id}/version/name/{urllib.parse.quote(version)}"
        try:
            response = self._auth._get(url)
            return ModelVersionOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "model version", version)
        except IndexError:
            raise ValueError("The model version is invalid. Please check the entered value.")

    def list_model_versions(
        self,
        model: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[ModelVersionOutput]:
        parent_model = ModelApi(self._auth).get_model(model, project, workspace)
        if isinstance(model, int):
            url = f"/metadata/project/{parent_model.project.id}/model/{model}/version"
        elif isinstance(model, str):
            url = f"/metadata/project/{parent_model.project.id}/model/{parent_model.id}/version"
        else:
            raise InvalidReferenceError("model", model)
        queries = {"index": page_index, "size": page_size}
        if search:
            queries["search"] = search
        model_versions = self._auth._get(url + "?" + urlencode(queries))
        return PagedResponse(
            item_cls=ModelVersionOutput,
            total=model_versions["total"],
            page=model_versions["page"],
            items=model_versions["items"],
        )

    def create_model_version(
        self,
        data: ModelVersionInput,
        model: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> ModelVersionOutput:
        parent_model = ModelApi(self._auth).get_model(model, project, workspace)
        if isinstance(model, int):
            url = f"/metadata/project/{parent_model.project.id}/model/{model}/version"
        elif isinstance(model, str):
            url = f"/metadata/project/{parent_model.project.id}/model/{parent_model.id}/version"
        else:
            raise InvalidReferenceError("model", model)
        response = self._auth._post(url, data)
        return ModelVersionOutput(**response)

    def update_model_version(
        self,
        data: ModelVersionInput,
        version: Reference,
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> ModelVersionOutput:
        model_version_object = self.get_model_version(version, model, project, workspace)
        url = f"/metadata/modelversion/{model_version_object.id}"
        try:
            response = self._auth._put(url, data)
            return ModelVersionOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePutHttpError(e, "dataset version", version)

    def delete_model_version(
        self,
        version: Reference,
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        try:
            model_version = self.get_model_version(version, model, project, workspace)
            self._auth._delete(f"/metadata/modelversion/{model_version.id}")
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "model version", version)

    def create_model_version_properties(
        self,
        version: Reference,
        properties: List[PropertyInput],
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        model_version = self.get_model_version(version, model, project, workspace)
        url = f"/metadata/project/{model_version['model']['project']['id']}/model/{model_version['model']['id']}/version/{model_version.id}/entityProperty"
        if properties:
            for property in properties:
                self._auth._post(
                    url,
                    property,
                )
        self._logger.info(
            f"Hyperparameters with names: {[k.key for k in properties]} successfully added to ModelVersion {model_version.name}."
        )

    def update_model_version_properties(
        self,
        version: Reference,
        property_id: int,
        properties: List[PropertyInput],
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        model_version = self.get_model_version(version, model, project, workspace)
        url = f"/metadata/project/{model_version['model']['project']['id']}/model/{model_version['model']['id']}/version/{model_version.id}/entityProperty/{property_id}"
        if properties:
            for property in properties:
                self._auth._put(
                    url,
                    property,
                )
        self._logger.info(
            f"Hyperparameters with names: {[k.key for k in properties]} successfully updated in ModelVersion {model_version.name}."
        )

    def create_model_version_metrics(
        self,
        version: Reference,
        metrics: List[MetricInput],
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> None:
        model_version = self.get_model_version(version, model, project, workspace)
        url = f"/metadata/project/{model_version['model']['project']['id']}/model/{model_version['model']['id']}/version/{model_version.id}/entityMetric"
        if metrics:
            for metric in metrics:
                self._auth._post(
                    url,
                    metric,
                )
        self._logger.info(
            f"Metrics with names: {[k.key for k in metrics]} successfully added to ModelVersion {model_version.name}."
        )

    def update_model_version_metrics(
        self,
        version: Reference,
        metric_id: int,
        metrics: List[MetricInput],
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        model_version = self.get_model_version(version, model, project, workspace)
        url = f"/metadata/project/{model_version['model']['project']['id']}/model/{model_version['model']['id']}/version/{model_version.id}/entityMetric/{metric_id}"
        if metrics:
            for metric in metrics:
                self._auth._put(
                    url,
                    metric,
                )
        self._logger.info(
            f"Metrics with names: {[k.key for k in metrics]} successfully updated in ModelVersion {model_version.name}."
        )

    def list_model_version_metrics(
        self,
        version: Reference,
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[MetricOutput]:
        model_version = self.get_model_version(version, model, project, workspace)
        url = f"/metadata/project/{model_version['model']['project']['id']}/model/{model_version['model']['id']}/version/{model_version.id}/entityMetric"
        queries = {"index": page_index, "size": page_size}
        if search is not None:
            queries["search"] = search
        model_version_metrics = self._auth._get(url + "?" + urlencode(queries))
        metrics = [MetricOutput.from_dict(metric).as_dict() for metric in model_version_metrics["items"]]
        return PagedResponse(
            item_cls=MetricOutput,
            total=model_version_metrics["total"],
            page=model_version_metrics["page"],
            items=metrics,
        )

    def list_model_version_properties(
        self,
        version: Reference,
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[PropertyOutput]:
        model_version = self.get_model_version(version, model, project, workspace)
        url = f"/metadata/project/{model_version['model']['project']['id']}/model/{model_version['model']['id']}/version/{model_version.id}/entityProperty"
        queries = {"index": page_index, "size": page_size}
        if search is not None:
            queries["search"] = search
        model_version_properties = self._auth._get(url + "?" + urlencode(queries))
        properties = [PropertyOutput.from_dict(property).as_dict() for property in model_version_properties["items"]]
        return PagedResponse(
            item_cls=PropertyOutput,
            total=model_version_properties["total"],
            page=model_version_properties["page"],
            items=properties,
        )

    def delete_model_version_metrics(
        self,
        version: Reference,
        metric_id: int,
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        model_version = self.get_model_version(version, model, project, workspace)
        url = f"/metadata/project/{model_version['model']['project']['id']}/model/{model_version['model']['id']}/version/{model_version.id}/entityMetric/{metric_id}"
        metric_list = self.list_model_version_metrics(model_version.id, model, project, workspace).list
        self._auth._delete(url)
        self._logger.info(
            f"Metrics with names: {[item.as_dict()['key']  for item in metric_list if item.as_dict()['id'] == metric_id]} successfully deleted from ModelVersion {model_version.name}."
        )

    def delete_model_version_properties(
        self,
        version: Reference,
        property_id: int,
        model: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ):
        model_version = self.get_model_version(version, model, project, workspace)
        url = f"/metadata/project/{model_version['model']['project']['id']}/model/{model_version['model']['id']}/version/{model_version.id}/entityProperty/{property_id}"
        hyperparameters_list = self.list_model_version_properties(model_version.id, model, project, workspace).list
        self._auth._delete(url)
        self._logger.info(
            f"Hyperparameters with names: {[item.as_dict()['key']  for item in hyperparameters_list if item.as_dict()['id'] == property_id]} successfully deleted from ModelVersion {model_version.name}."
        )



from __future__ import annotations

from typing import Optional, Union, List, Dict, BinaryIO, TYPE_CHECKING, Tuple, Any

from .attachment_container import AttachmentContainer
from ..api.json import ArtifactVersion, FileMetadata

if TYPE_CHECKING:
    from vectice.models import Dataset


class DatasetVersion(AttachmentContainer):
    """
    Describes a dataset version
    """

    def __init__(
        self,
        dataset: Dataset,
        id: int,
        name: str,
        version_number: int,
        properties: Optional[Dict[str, str]] = None,
        description: Optional[str] = None,
        resources: Optional[Union[str, List[str]]] = None,
        attachments: Optional[Union[str, List[str]]] = None,
        auto_version: Optional[bool] = True,
        is_starred: Optional[bool] = False,
        version: Optional[ArtifactVersion] = None,
    ):
        """
        :param dataset: The parent Dataset object.
        :param id: The project identifier.
        :param name: The name of the dataset.
        :param version_number: The version number of the dataset version.
        :param properties: The dataset version properties.
        :param description: A quick description of the dataset.
        :param resources: The dataset version resources.
        :param attachments: The dataset version attachments.
        :param auto_version: Boolean for auto versioning option.
        :param is_starred: Boolean for starring the dataset version.
        :param version: The artifact version object of the dataset version.
        """
        super().__init__(name, id, dataset._client, "DatasetVersion")
        self._dataset = dataset
        self._description = description
        self._isStarred = is_starred
        self._autoVersion = auto_version
        self._properties = properties
        self._version = version  # Relevant for runs // ArtifactVersion
        self._version_number = version_number
        self._resources = resources
        self._parentId = dataset.id
        self._attachments: Optional[
            Union[Tuple[str, Tuple[str, BinaryIO]], List[Tuple[str, Tuple[str, BinaryIO]]]]
        ] = None

        if properties:
            self.properties = properties
        if version:
            if isinstance(version, str):
                self.version = ArtifactVersion(versionNumber=None, versionName=version, id=None)
            elif isinstance(version, int):
                self.version = ArtifactVersion(versionNumber=version, versionName=None, id=None)
        if attachments:
            self.add_attachments(attachments)

    def __repr__(self):
        return f"DatasetVersion(dataset={self.dataset}, id={self.id}, description={self.description}, is_starred={self.is_starred}, auto_version={self.auto_version}, name={self.name}, properties={self.properties}, version={self.version})"

    @property
    def dataset(self) -> Dataset:
        """
        Parent Dataset object.
        :return: Dataset
        """
        return self._dataset

    @property
    def id(self) -> int:
        """
        Dataset version identifier.
        :return: int
        """
        return self._id

    @id.setter
    def id(self, dataset_version_id: int):
        self._id = dataset_version_id

    @property
    def name(self) -> Optional[str]:
        """
        Dataset version name.
        :return: Optional[str]
        """
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def parent_name(self) -> Optional[str]:
        """
        Parent dataset name.
        :return: Optional[str]
        """
        return self._dataset.name

    @property
    def parent_id(self) -> Optional[int]:
        """
        Parent dataset id.
        :return: Optional[int]
        """
        return self._dataset.id

    @parent_id.setter
    def parent_id(self, dataset: int):
        self._parentId = dataset

    @property
    def version_number(self) -> int:
        """
        Version number.
        :return: int
        """
        return self._version_number

    @property
    def description(self) -> Optional[str]:
        """
        Quick description of the dataset version
        :return: Optional[str]
        """
        return self._description

    @description.setter
    def description(self, description: str):
        self._description = description

    @property
    def properties(self) -> Optional[Dict]:
        """
        Dataset version properties.
        :return: Optional[Dict]
        """
        return self._properties

    @properties.setter
    def properties(self, properties: Dict[str, Any]):
        self._properties = properties

    @property
    def resources(self) -> Optional[Union[str, List[str]]]:
        """
        Dataset version resources.
        :return: Optional[Union[str, List[str]]]
        """
        return self._resources

    @resources.setter
    def resources(self, resources: Union[str, List[str]]):
        self._resources = resources

    @property
    def is_starred(self) -> Optional[bool]:
        """
        Boolean for the dataset version starring.
        :return: Optional[bool]
        """
        return self._isStarred

    @is_starred.setter
    def is_starred(self, is_starred: bool):
        self._isStarred = is_starred

    @property
    def auto_version(self) -> Optional[bool]:
        """
        Boolean for auto versioning option.
        :return: Optional[bool]
        """
        return self._autoVersion

    @auto_version.setter
    def auto_version(self, auto_version: bool):
        self._autoVersion = auto_version

    @property
    def version(self) -> Optional[ArtifactVersion]:
        """
        The artifact version.
        :return: Optional[ArtifactVersion]
        """
        return self._version

    @version.setter
    def version(self, version: ArtifactVersion):
        self._version = version

    @property
    def attachments(self) -> Optional[Union[Tuple[str, Tuple[str, BinaryIO]], List[Tuple[str, Tuple[str, BinaryIO]]]]]:
        """
        The attachments linked to the dataset version.
        """
        return self._attachments

    @attachments.setter
    def attachments(self, attachments: Union[Tuple[str, Tuple[str, BinaryIO]], List[Tuple[str, Tuple[str, BinaryIO]]]]):
        self._attachments = attachments

    def list_files_metadata(self) -> List[FileMetadata]:
        """
        Lists the dataset version files metadata.
        """
        return self._client.list_dataset_version_files_metadata(self.id)



from __future__ import annotations

import logging
import datetime
from typing import Optional, Dict, List, Union, TYPE_CHECKING, Any


from vectice.api import Client
from vectice.api.json import DatasetVersionInput, FileMetadata, VersionStrategy
from vectice.api.reference import InvalidReferenceError
from .dataset_version import DatasetVersion
from .data_resource import DataResource

if TYPE_CHECKING:
    from vectice import Reference
    from vectice.models import Project, Connection


class Dataset:
    """
    Describe a dataset

    There is several type of dataset:

    - dataset based on a Connection.
    - dataset without any Connection where we store only metadata
    - dataset without any Connection where we store local files and the associated metadata

    For dataset based on a Connection, The platform store metadata of the current state of the shared dataset.
    The SDK does not manage the metadata.

    For dataset without any Connection, the SDK calculate the metadata from the current state of the dataset.


    """

    def __init__(
        self,
        name: str,
        id: int,
        project: Project,
        description: Optional[str] = None,
        connection: Optional[Connection] = None,
        pattern: Optional[str] = "*",
        resources: Optional[List[str]] = None,
    ):
        """
        :param name: the name of the dataset
        :param id: the project identifier
        :param project: the project the dataset belong to
        :param description: a quick description of the dataset
        :param connection: the connection the dataset is linked to
        :param pattern: the pattern of the dataset
        """
        self._id = id
        self._project: Project = project
        self._name = name
        self._description = description
        self._connection = connection
        self._resources: Optional[List[str]] = resources
        self._pattern = pattern
        self._client: Client = project._client
        self._logger = logging.getLogger(self.__class__.__name__)

    def __repr__(self):
        return f"Dataset(name={self.name}, id={self.id}, description={self.description}, connection={self.connection}, resources={self.resources})"

    @property
    def id(self) -> int:
        """
        Dataset identifier.
        :return: int
        """
        return self._id

    @id.setter
    def id(self, dataset_id: int):
        self._id = dataset_id

    @property
    def project(self) -> Optional[Project]:
        """
        The project this dataset belong to.
        :return: Optional[Project]
        """
        return self._project

    @property
    def name(self) -> str:
        """
        Name of the dataset
        :return: str
        """
        return self._name

    @property
    def description(self) -> Optional[str]:
        """
        Quick description of the dataset
        :return: Optional[str]
        """
        return self._description

    @description.setter
    def description(self, description: str):
        self._description = description

    @property
    def connection(self) -> Optional[Connection]:
        """
        the connection used to retrieve this dataset
        :return: Optional[Connection]
        """
        return self._connection

    @property
    def resources(self) -> Optional[List[str]]:
        """
        the connection used to retrieve this dataset
        :return: Optional[List[str]]
        """
        return self._resources

    @resources.setter
    def resources(self, resources: List[str]):
        self._resources = resources

    def get_dataset_version(
        self,
        version: Reference,
    ) -> DatasetVersion:
        """
        Gets a specific version of this dataset.

        :param version: The version name or id

        :return: A DatasetVersion or None if the version does not exist
        """
        dataset_version_output = self._client.get_dataset_version(version, self._id)
        try:
            properties: Dict = {
                prop.key: prop.value for prop in self._client.list_dataset_version_properties(version, self._id).list
            }
        except Exception as e:
            logging.warning(f"Properties were not found due to {e}")
            properties = {}
        if dataset_version_output is not None:
            self._logger.info(
                f"DatasetVersion with id: {dataset_version_output.id} successfully retrieved from Dataset {self.name}."
            )
            return DatasetVersion(
                dataset=self,
                id=dataset_version_output.id,
                name=dataset_version_output.name,
                version_number=dataset_version_output.version_number,
                description=dataset_version_output.description,
                properties=properties,
                is_starred=dataset_version_output.is_starred,
            )
        else:
            raise RuntimeError(f"The Dataset Version {version} was not found.")

    def create_dataset_version(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_starred: Optional[bool] = None,
        properties: Optional[Dict[str, Any]] = None,
        resources: Optional[List[str]] = None,
        metadata: Optional[List[FileMetadata]] = None,
        attachments: Optional[Union[str, List[str]]] = None,
        version_strategy: VersionStrategy = VersionStrategy.MANUAL,
    ) -> DatasetVersion:
        """
        Creates a new dataset version

        :param name: The name of the dataset version e.g Changes 'Version 1' to 'name'
        :param description: The description of the dataset version
        :param is_starred: Whether the dataset version is starred
        :param properties: The properties of the dataset version
        :param resources: if connection is set, resources present in connection that should be added to dataset
        :param metadata: if no connection is set, you can add raw metadata to allow versioning of dataset
        :param attachments: list of files to be added to the dataset version (not part of the dataset)

        :return: A DatasetVersion
        """

        dataset_version_input = DatasetVersionInput(name=name, description=description, isStarred=is_starred)
        if resources is not None:
            dataset_version_input["resources"] = [DataResource.create_resource(uri) for uri in resources]
        if metadata is not None:
            dataset_version_input["filesMetadata"] = metadata
        if properties is not None:
            property_list = [dict(key=k, value=v) for k, v in properties.items()]
            dataset_version_input["properties"] = property_list

        dataset_version_input["autoVersion"] = version_strategy == VersionStrategy.AUTOMATIC
        dataset_version_output = self._client.create_dataset_version(dataset_version_input, self._id)
        result = DatasetVersion(
            dataset=self,
            id=dataset_version_output.id,
            name=dataset_version_output.name,
            version_number=dataset_version_output.version,
            properties=properties,
            description=dataset_version_output.description,
            is_starred=dataset_version_output.is_starred,
            resources=resources,
        )
        if attachments is not None:
            result.add_attachments(attachments)
        if dataset_version_output.reused_version:
            self._logger.info(f"DatasetVersion with id: {result.id} successfully reused in Dataset {self.name}.")
        else:
            self._logger.info(f"DatasetVersion with id: {result.id} successfully created in Dataset {self.name}.")
        return result

    def update_dataset_version(
        self,
        version: Reference,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_starred: Optional[bool] = None,
        properties: Optional[Dict] = None,
        resources: Optional[List[str]] = None,
        metadata: Optional[List[FileMetadata]] = None,
    ) -> DatasetVersion:
        """
        Update a dataset version.

        :param version: The dataset version name or id
        :param name: The name of the dataset version e.g Changes 'Version 1' to 'name'
        :param description: The description of the dataset version
        :param is_starred: Whether the dataset version is starred
        :param properties: The properties of the dataset version
        :param resources: if connection is set, resources present in connection that should be added to dataset
        :param metadata: if no connection is set, you can add raw metadata to allow versioning of dataset

        :return: DatasetVersion
        """
        properties_list = self._client.list_dataset_version_properties(version, self._id).list
        for prop in properties_list:
            if properties and (properties.get(prop.key) is not None):
                self._client.update_dataset_version_properties(
                    version,
                    property_id=prop.id,
                    properties={
                        "key": prop.key,
                        "value": properties[prop.key],
                        "timestamp": datetime.datetime.now().isoformat(),
                    },
                    dataset=self._id,
                )
                properties.pop(prop.key)
        if properties:
            for key in properties:
                self._client.create_dataset_version_properties(
                    version,
                    properties={"key": key, "value": properties[key], "timestamp": datetime.datetime.now().isoformat()},
                    dataset=self._id,
                )
        dataset_version_output = self._client.get_dataset_version(version, self._id)
        if dataset_version_output is None:
            raise InvalidReferenceError("version", version)
        if resources is not None:
            dataset_version_output["resources"] = [DataResource.create_resource(item) for item in resources]
        if metadata is not None:
            dataset_version_output["filesMetadata"] = metadata
        if name is not None:
            dataset_version_output["name"] = name
        if description is not None:
            dataset_version_output["description"] = description
        if is_starred is not None:
            dataset_version_output["isStarred"] = is_starred
        new_properties: Dict = {
            prop.key: prop.value for prop in self._client.list_dataset_version_properties(version, self._id).list
        }
        dataset_version_output = self._client.update_dataset_version(dataset_version_output, version, self._id)
        self._logger.info(
            f"DatasetVersion with id: {dataset_version_output.id} successfully updated in Dataset {self.name}."
        )
        return DatasetVersion(
            dataset=self,
            id=dataset_version_output.id,
            name=dataset_version_output.name,
            version_number=dataset_version_output.version_number,
            properties=new_properties,
            description=dataset_version_output.description,
            is_starred=dataset_version_output.is_starred,
            resources=dataset_version_output.get("resources"),
        )

    def list_dataset_versions(
        self, search: Optional[str] = None, page_index: int = 1, page_size: int = 20
    ) -> List[DatasetVersion]:
        """
        List the dataset versions.

        Returns a List of the Dataset Versions associated with the Dataset.
        The DatasetVersion can be used in runs or experiments.

        :param search: The name to search
        :param page_index: The page index
        :param page_size: The page size

        :return: List of DatasetVersions
        """
        dataset_version_output_list = self._client.list_dataset_versions(self._id, search, page_index, page_size)
        return [
            DatasetVersion(
                dataset=self,
                id=dataset_version_output.id,
                name=dataset_version_output.name,
                version_number=dataset_version_output.version_number,
                description=dataset_version_output.description,
                is_starred=dataset_version_output.is_starred,
            )
            for dataset_version_output in dataset_version_output_list.list
        ]

    def delete_dataset_version(
        self,
        version: Reference,
    ) -> None:
        """
        Delete a specific version of this dataset.

        :param version: The dataset version name or id
        :return: None
        """
        result = self._client.get_dataset_version(version, self._id)
        self._client.delete_dataset_version(version, self._id)
        self._logger.info(f"DatasetVersion with id: {result.id} successfully deleted from Dataset {self.name}.")



from datetime import datetime, timezone
from typing import Optional, Union


class Metric:
    def __init__(self, key, value, name=None):
        self._key: str = key
        """"""
        self._value: str = value
        """"""
        self._timestamp: Union[datetime, str] = datetime.now(timezone.utc).isoformat()
        """"""
        self._name: Optional[str] = None
        """"""

    @property
    def key(self):
        return self._key

    @property
    def value(self):
        return self._value

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def name(self):
        return self._name



from dataclasses import dataclass

from typing import Optional


@dataclass
class AttachmentOutput:
    fileId: int
    fileName: str
    contentType: str
    entityId: int
    entityType: str
    modelFramework: Optional[str] = None



from __future__ import annotations

import logging
from datetime import datetime
from typing import Optional, List, TYPE_CHECKING, Union, Dict, Tuple

from vectice.api.json import (
    RunStatus,
    ArtifactType,
    StopRunInput,
    ArtifactVersion,
    ArtifactReferenceInput,
    RulesDatasetVersionInput,
    RulesModelVersionInput,
    RulesCodeVersionInput,
    RunInput,
    JobInput,
    StartRunInput,
    ArtifactReferenceOutput,
    VersionStrategy,
)
from vectice.models import CodeVersion
from vectice.models import DatasetVersion
from vectice.models import ModelVersion
from .artifact_reference import ArtifactReference
from .attachment_container import AttachmentContainer
from .git_version import GitVersion
from .integration import AbstractIntegration

if TYPE_CHECKING:
    from vectice.models import Job, Stage


logger = logging.getLogger(__name__)


HEADERS = {
    "dataset": "The Dataset Used:",
    "datasets": "The Datasets Used:",
    "model": "The Model Used:",
    "models": "The Models Used:",
    "modelversion": "The Model Version Used:",
    "modelversions": "The Model Versions Used:",
    "mixed": "The Models And Datasets Used:",
    "dataset output": "The Output Dataset:",
    "datasets output": "The Output Datasets:",
    "model output": "The Output Model:",
    "models output": "The Output Models:",
    "modelversion output": "The Output Model Version:",
    "modelversions output": "The Output Model Versions:",
    "mixed output": "The Output Models And Datasets:",
}


def select_header(items, input=False, output=False) -> Tuple[str, bool]:
    datasets, dataset_versions, models, model_versions = 0, 0, 0, 0

    for item in items:
        if (
            (item.artifact_type.name == "DATASET" and item.version_id)
            or (item.artifact_type.name == "DATASET" and item.version_name)
            or (item.artifact_type.name == "DATASET" and item.version_strategy.value == "AUTOMATIC")
        ):
            dataset_versions += 1
        elif item.artifact_type.name == "DATASET":
            datasets += 1
        elif (item.artifact_type.name == "MODEL" and item.version_id) or (
            item.artifact_type.name == "MODEL" and item.version_name
        ):
            model_versions += 1
        elif item.artifact_type.name == "MODEL":
            models += 1

    if (datasets == 1 or dataset_versions == 1) and models == 0 and model_versions == 0 and input is True:
        return HEADERS["dataset"], False

    elif (datasets >= 1 or dataset_versions >= 1) and models == 0 and model_versions == 0 and input is True:
        return HEADERS["datasets"], True

    elif datasets == 0 and models == 1 and model_versions == 0 and dataset_versions == 0 and input is True:
        return HEADERS["model"], False

    elif datasets == 0 and models >= 1 and model_versions == 0 and dataset_versions == 0 and input is True:
        return HEADERS["models"], True

    elif datasets == 0 and models == 0 and model_versions == 1 and dataset_versions == 0 and input is True:
        return HEADERS["modelversion"], False

    elif datasets == 0 and (models >= 1 or model_versions >= 1) and dataset_versions == 0 and input is True:
        return HEADERS["models"], True

    elif (
        (input is True and datasets >= 1 and models >= 1)
        or (input is True and model_versions >= 1 and dataset_versions >= 1)
        or (input is True and datasets >= 1 and model_versions >= 1)
        or (input is True and models >= 1 and dataset_versions >= 1)
    ):
        return HEADERS["mixed"], True

    elif (datasets == 1 or dataset_versions == 1) and models == 0 and model_versions == 0 and output is True:
        return HEADERS["dataset output"], False

    elif (datasets >= 1 or dataset_versions >= 1) and models == 0 and model_versions == 0 and output is True:
        return HEADERS["datasets output"], True

    elif datasets == 0 and models == 1 and model_versions == 0 and dataset_versions == 0 and output is True:
        return HEADERS["model output"], False

    elif datasets == 0 and models >= 1 and model_versions == 0 and dataset_versions == 0 and output is True:
        return HEADERS["models output"], True

    elif datasets == 0 and models == 0 and model_versions == 1 and dataset_versions == 0 and output is True:
        return HEADERS["modelversion output"], False

    elif datasets == 0 and models == 0 and model_versions >= 1 and dataset_versions == 0 and output is True:
        return HEADERS["modelversions output"], True

    elif datasets == 0 and (models >= 1 or model_versions >= 1) and dataset_versions == 0 and output is True:
        return HEADERS["models output"], True

    elif (
        (output is True and datasets >= 1 and models >= 1)
        or (output is True and model_versions >= 1 and dataset_versions >= 1)
        or (output is True and datasets >= 1 and model_versions >= 1)
        or (output is True and models >= 1 and dataset_versions >= 1)
    ):
        return HEADERS["mixed output"], True

    else:
        raise ValueError("No header was assigned.")


def inputs_documentation(inputs: List[ArtifactReference], stage: Stage, client) -> None:
    if len(inputs) < 1:
        logging.warning("There are no inputs for the run.")
        return
    header, multiple = select_header(inputs, input=True)
    if multiple is True:
        stage._add_header(header)
        stage.add_block(text=" ")
    for item in inputs:
        if item.artifact_type.name == "DATASET":
            if multiple is False:
                stage._add_header(header)
                stage.add_block(text=" ")
            if item.version_id:
                if item.description:
                    stage.add_block(text=item.description)
                stage.add_block(dataset_version=item.version_id)
            elif item.version_strategy.value == "AUTOMATIC":
                dataset_versions: List[DatasetVersion] = sorted(
                    client.list_dataset_versions(item.dataset).list, key=lambda x: x.id
                )
                stage.add_block(dataset_version=dataset_versions[-1].id)
            elif item.dataset and isinstance(item.dataset, int):
                stage.add_block(dataset=item.dataset)
        elif item.artifact_type.name == "MODEL":
            if multiple is False:
                stage._add_header(header)
                stage.add_block(text=" ")
            if item.description:
                stage.add_block(text=item.description)
            if item.version_id:
                attachments = client.list_attachments("modelversion", item.version_id).list
                if item.model:
                    stage._add_model_table(item.model, item.version_id)
                if len(attachments) >= 1:
                    stage.add_block(text="The Model Version Attachments:")
                    for attachment in attachments:
                        stage._add_attachment(attachment)
            elif item.model and isinstance(item.model, int):
                stage.add_block(model=item.model)


def outputs_documentation(outputs: List[ArtifactReference], stage: Stage, client):
    if len(outputs) < 1:
        logging.warning("There are no outputs for the run.")
        return
    header, multiple = select_header(outputs, output=True)
    if multiple is True:
        stage._add_header(header)
        stage.add_block(text=" ")
    for item in outputs:
        if item.artifact_type.name == "MODEL":
            if multiple is False:
                stage._add_header(header)
                stage.add_block(text=" ")
            if item.description:
                stage.add_block(text=item.description)
            if item.version_id:
                attachments = client.list_attachments("modelversion", item.version_id).list
                if item.model:
                    stage._add_model_table(item.model, item.version_id)
                if len(attachments) >= 1:
                    stage.add_block(text="The Model Version Attachments:")
                    for attachment in attachments:
                        stage._add_attachment(attachment)
            elif item.model and isinstance(item.model, int):
                stage.add_block(model=item.model)
        elif item.artifact_type.name == "DATASET":
            if multiple is False:
                stage._add_header(header)
                stage.add_block(text=" ")
            if item.description:
                stage.add_block(text=item.description)
            if item.version_id:
                stage.add_block(dataset_version=item.version_id)
            elif item.version_strategy.value == "AUTOMATIC":
                dataset_versions: List[DatasetVersion] = sorted(
                    client.list_dataset_versions(item.dataset).list, key=lambda x: x.id
                )
                stage.add_block(dataset_version=dataset_versions[-1].id)
            elif item.dataset and isinstance(item.dataset, int):
                stage.add_block(dataset=item.dataset)


def __create_artifact_input__(item: ArtifactReference) -> ArtifactReferenceInput:
    result = ArtifactReferenceInput(item.artifact_type)
    if item.dataset is not None:
        result.dataset = __to_dataset_version_input__(item)
    elif item.model is not None:
        result.model = __to_model_version_input__(item)
    elif item.code is not None:
        result.code = __to_code_version_input__(item)
    return result


def __to_code_version_input__(artifact_reference: ArtifactReference) -> RulesCodeVersionInput:
    result = RulesCodeVersionInput()
    if isinstance(artifact_reference.code, int):
        result.parentId = artifact_reference.code
    else:
        result.parentName = str(artifact_reference.code)
    result.version = ArtifactVersion(version_id=artifact_reference.version_id)
    return result


def __to_dataset_version_input__(artifact_reference: ArtifactReference) -> RulesDatasetVersionInput:
    result = RulesDatasetVersionInput()
    if isinstance(artifact_reference.dataset, int):
        result.parentId = artifact_reference.dataset
    else:
        result.parentName = str(artifact_reference.dataset)
    if artifact_reference.description is not None:
        result.description = artifact_reference.description
    if artifact_reference.version_id is not None:
        result.autoVersion = False
        result.version = ArtifactVersion(version_id=artifact_reference.version_id)
        return result
    if artifact_reference.version_name is not None:
        result.autoVersion = False
        result.version = ArtifactVersion(version_name=artifact_reference.version_name)
        return result
    if artifact_reference.version_number is not None:
        result.autoVersion = False
        result.version = ArtifactVersion(version_number=artifact_reference.version_number)
        return result
    if artifact_reference.version_strategy is not None:
        result.autoVersion = True
        return result
    raise RuntimeError(f"Missing version information for dataset {artifact_reference.dataset}")


def __to_model_version_input__(artifact_reference: ArtifactReference) -> RulesModelVersionInput:
    result = RulesModelVersionInput()
    if isinstance(artifact_reference.model, int):
        result.parentId = artifact_reference.model
    else:
        result.parentName = str(artifact_reference.model)
    if artifact_reference.description is not None:

        result.description = artifact_reference.description
    if artifact_reference.version_id is not None:
        result.version = ArtifactVersion(id=artifact_reference.version_id)
        return result
    if artifact_reference.version_name is not None:
        result.version = ArtifactVersion(version_name=artifact_reference.version_name)
        return result
    if artifact_reference.version_number is not None:
        result.version = ArtifactVersion(version_number=artifact_reference.version_number)
        return result
    raise RuntimeError(f"Missing version information for model {artifact_reference.model}")


def _ensure_reference(item: Union[ArtifactReference, CodeVersion, ModelVersion, DatasetVersion]) -> ArtifactReference:
    if isinstance(item, ArtifactReference):
        return item
    elif isinstance(item, CodeVersion):
        return ArtifactReference(code=item.code_id, version_id=item.id)
    elif isinstance(item, DatasetVersion):
        return ArtifactReference(dataset=item.dataset.id, version_id=item.id, description=item.description)
    elif isinstance(item, ModelVersion):
        return ArtifactReference(model=item.model.id, version_id=item.id, description=item.description)
    else:
        raise RuntimeError(f"can not create artifact reference from unknown object '{item}'")


class Run(AttachmentContainer):
    """
    The Run acts as the ActiveRun, JobRun and Job.
    The run is not viewable in the Vectice platform until the job is started.
    """

    def __init__(
        self,
        id: int,
        job: Job,
        name: str,
        system_name: Optional[str] = "",
        created_date: Optional[datetime] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        status: RunStatus = RunStatus.SCHEDULED,
        auto_code: bool = False,
        check_remote_repository: bool = True,
        description: Optional[str] = None,
        properties: Optional[Dict[str, str]] = None,
    ):
        """
        :param id: The id of the run
        :param job: The job of the run
        :param name: The name of the run
        :param system_name: The system name this run was executed in
        :param created_date: The creation date of the run
        :param start_date: The starting date of the run
        :param end_date: The ending date of the run
        :param status: The status of the run
        :param auto_code: If the auto code is activated
        :param check_remote_repository: If checking remote repository option is activated.
        :param description: A quick description of the run
        :param properties: The properties of the run
        """
        super().__init__(name, id, job._client, "Run")
        self._job = job
        self._systemName = system_name
        self._description = description
        self._properties = properties
        self._createdDate = created_date
        self._startDate = start_date
        self._endDate = end_date
        self._status = status
        self._duration: Optional[int] = None
        self._integration_client: Optional[AbstractIntegration] = job._integration_client
        self._inputs: List[ArtifactReference] = []
        self._outputs: List[ArtifactReference] = []
        self._auto_code = auto_code
        self._check_remote_repository = check_remote_repository
        self._auto_document_run = False
        self._document_stage = None
        self._logger = logging.getLogger(self.__class__.__name__)

    def __repr__(self):
        return f"Run(id={self.id}, job={self.job}, name={self.name}, system_name={self.system_name}, start_date={self.start_date}, end_date={self.end_date}, duration={self.duration}, status={self.status}, inputs={self.inputs}, outputs={self.outputs})"

    def __enter__(self) -> Run:
        if self._status == RunStatus.SCHEDULED:
            self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        status = RunStatus.COMPLETED if exc_type is None else RunStatus.FAILED
        try:
            self.end_run(status=status)
        except Exception as e:
            self.fail(reason=str(e))
        finally:
            return exc_type is None

    @property
    def id(self) -> int:
        """
        The run identifier.
        :return: int
        """
        return self._id

    @id.setter
    def id(self, id: int):
        self._id = id

    @property
    def job(self) -> Job:
        """
        The parent Job object of the run.
        :return: Job
        """
        return self._job

    @job.setter
    def job(self, job):
        self._job = job

    @property
    def name(self) -> str:
        """
        The run name.
        :return: str
        """
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def system_name(self) -> Optional[str]:
        """
        The system name the run is executed in.
        :return: Optional[str]
        """
        return self._systemName

    @system_name.setter
    def system_name(self, system_name: str):
        self._systemName = system_name

    @property
    def description(self) -> Optional[str]:
        """
        A quick description of the run.
        :return: Optional[str]
        """
        return self._description

    @description.setter
    def description(self, description: str):
        self._description = description

    @property
    def properties(self) -> Optional[Dict]:
        """
        The run properties.
        :return: Optional[Dict]
        """
        return self._properties

    @properties.setter
    def properties(self, properties: Dict):
        self._properties = properties

    @property
    def start_date(self) -> Optional[datetime]:
        """
        The starting date of the run.
        :return: Optional[datetime]
        """
        return self._startDate

    @property
    def end_date(self) -> Optional[datetime]:
        """
        The ending date of the run.
        :return: Optional[datetime]
        """
        return self._endDate

    @end_date.setter
    def end_date(self, end_date: datetime):
        self._endDate = end_date

    @property
    def status(self) -> RunStatus:
        """
        The status of the run.
        :return: RunStatus
        """
        return self._status

    @status.setter
    def status(self, status: RunStatus):
        self._status = status

    @property
    def duration(self) -> Optional[int]:
        """
        The duration of the run.
        :return: Optional[int]
        """
        return self._duration

    @duration.setter
    def duration(self, duration: int):
        self._duration = duration

    @property
    def inputs(self) -> List[ArtifactReference]:
        """
        The artifact inputs of the run.
        :return: List[ArtifactReference]
        """
        return self._inputs.copy()

    @property
    def outputs(self) -> List[ArtifactReference]:
        """
        The artifact outputs of the run.
        :return: List[ArtifactReference]
        """
        return self._outputs.copy()

    def to_reference_input(self, item: ArtifactReference) -> ArtifactReferenceInput:
        return ArtifactReferenceInput(
            item.artifact_type,
            item.description,
            RulesDatasetVersionInput(
                parentName=str(item.dataset) if not isinstance(item.dataset, int) else None,
                parentId=item.dataset if isinstance(item.dataset, int) else None,
                version=ArtifactVersion(item.version_number, item.version_name, item.version_id)
                if item.version_strategy == VersionStrategy.MANUAL
                else None,
                autoVersion=True if item.version_strategy == VersionStrategy.AUTOMATIC else None,
                description=item.description,
            )
            if item.artifact_type == ArtifactType.DATASET
            else None,
            RulesModelVersionInput(
                parentName=str(item.model) if not isinstance(item.model, int) else None,
                parentId=item.model if isinstance(item.model, int) else None,
                version=ArtifactVersion(item.version_number, item.version_name, item.version_id),
                description=item.description,
            )
            if item.artifact_type == ArtifactType.MODEL
            else None,
            RulesCodeVersionInput(
                parentName=str(item.model) if not isinstance(item.model, int) else None,
                parentId=item.model if isinstance(item.model, int) else None,
                version=ArtifactVersion(item.version_number, item.version_name, item.version_id),
            )
            if item.artifact_type == ArtifactType.CODE
            else None,
        )

    def add_output(self, output: ArtifactReference):
        """
        Adds an artifact as output to the run.

        :param output: The artifact to be added as an output
        :return: None
        """
        if self._outputs is None:
            self._outputs = [output]
        else:
            self._outputs.append(output)
        if self.status != RunStatus.SCHEDULED:
            self._client.fill_run(
                self.id,
                outputs=[self.to_reference_input(output)],
            )

    def add_outputs(self, outputs: List[ArtifactReference]):
        """
        Adds a list of artifacts as outputs to the run.

        :param outputs: The list of artifacts to be added as outputs
        :return: None
        """
        if self._outputs is None:
            self._outputs = outputs
        else:
            self._outputs.extend(outputs)
        if self.status != RunStatus.SCHEDULED:
            self._client.fill_run(
                self.id,
                outputs=[self.to_reference_input(output) for output in outputs],
            )

    def add_input(self, input: ArtifactReference):
        """
        Adds an artifact as input to the run.

        :param input: The artifact to be added as an input
        :return: None
        """
        if self._inputs is None:
            self._inputs = [input]
        else:
            if input not in self._inputs:
                self._inputs.append(input)
                # finally, if run is started, add it to the server
                if self.status != RunStatus.SCHEDULED:
                    response = self._client.fill_run(
                        self.id,
                        inputs=[self.to_reference_input(input)],
                    )
                    # artifacts = [ArtifactReferenceOutput(**item) for item in response["outputArtifacts"]]
                    # reused_dataset = [
                    #     item for item in artifacts if item["reusedVersion"] is True and item.get("dataSetVersion")
                    # ]
                    # if len(reused_dataset) >= 1:
                    #     for dataset_version in reused_dataset:
                    #         parent_dataset = dataset_version["dataSetVersion"]["dataSet"]["name"]
                    #         self._logger.info(
                    #             f"Reusing version {dataset_version['dataSetVersionId']} of dataset {parent_dataset}"
                    #         )
                    artifacts = [ArtifactReferenceOutput(**item) for item in response["outputArtifacts"]]
                    for artifact in artifacts:
                        # logging.info(artifact)
                        if artifact["reusedVersion"] is True:
                            # self._logger.info(artifact)
                            parent_dataset = self._client.get_dataset(artifact["dataSetVersion"]["dataSetId"])
                            # self._logger.info(artifact["dataSetVersion"]["dataSetId"])
                            self._logger.info(
                                f"Reusing version {artifact['dataSetVersionId']} of dataset {parent_dataset.name}"
                            )
                        else:
                            logging.info("ICIIIIIIIII")
                            if (
                                artifact.get("dataSetVersion", None) is not None
                                and artifact["dataSetVersion"]["versionType"] is True
                            ):
                                parent_dataset = self._client.get_dataset(artifact["dataSetVersion"]["dataSetId"])
                                self._logger.info(
                                    f"Changes detected automatically in resources for dataset {parent_dataset.name}"
                                )
                                self._logger.info(f"Creating and using version with id: {artifact['dataSetVersionId']}")

    def add_inputs(self, artifacts: List[ArtifactReference]):
        """
        Adds a list of artifact as inputs to the run.

        :param artifacts: The list of artifacts to be added as inputs
        :return: None
        """
        # first, remove duplicate from provided list
        inputs = []
        for input in artifacts:
            if input not in inputs:
                inputs.append(input)
        # then call add_input
        for input in inputs:
            self.add_input(input)

    def start(
        self,
        inputs: Optional[List[Union[ArtifactReference, CodeVersion, ModelVersion, DatasetVersion]]] = None,
        auto_code: Optional[bool] = None,
        check_remote_repository: Optional[bool] = None,
    ) -> None:
        """
        Starts the run.
        :param inputs: The artifacts to be added as inputs to the run
        :param auto_code: Boolean for the auto code option
        :param check_remote_repository: Boolean for the checking remote repository option
        :return: None
        """
        if inputs is not None:
            artifacts_references_inputs = [_ensure_reference(item) for item in inputs if item is not None]
            self.add_inputs(artifacts_references_inputs)

        if auto_code is True or (auto_code is None and self._auto_code is True):
            code_artifact_is_present = False
            if self._inputs is not None:
                for artifact in self._inputs:
                    code_artifact_is_present = code_artifact_is_present or artifact.artifact_type == ArtifactType.CODE
            if not code_artifact_is_present:
                git_version = GitVersion.create(
                    check_remote_repository=check_remote_repository
                    if check_remote_repository is not None
                    else self._check_remote_repository
                )
                if git_version:
                    code_version = self.job.project.create_code_version(git_version=git_version)
                    self.add_input(ArtifactReference(code=code_version.code_id, version_id=code_version.id))

        job_input = JobInput(self.job.name, self.job.description, self.job.type)
        self._status = RunStatus.STARTED
        self._startDate = datetime.utcnow()
        run_input = RunInput(self.name, job_input, self.system_name, self._startDate, self.end_date, self.status)
        if self._inputs is None:
            artifact_inputs = None
        else:
            artifact_inputs = [__create_artifact_input__(artifact) for artifact in self._inputs]
        rule_input = StartRunInput(job_input, run_input, artifact_inputs)
        if self._integration_client:
            try:
                self._integration_client.before_start(self)
            except Exception as e:
                logging.warning(e)
        self._client.start_run(rule_input)
        if self._integration_client:
            try:
                self._integration_client.after_start(self)
            except Exception as e:
                logging.warning(e)

    def end_run(
        self,
        outputs: Optional[List[Union[ArtifactReference, CodeVersion, ModelVersion, DatasetVersion]]] = None,
        status: RunStatus = RunStatus.COMPLETED,
        reason: Optional[str] = None,
    ) -> None:
        """
        Ends the current (last) active run started by :func:`~Vectice.start_run`.
        To end a specific run, use :func:`~Vectice.stop_run` instead.

        :param outputs: The artifacts to be added as outputs to the run
        :param status: Status of the run changes by default to COMPLETED
        :param reason: Reason note for ending the run
        :return: None
        """
        if outputs is not None:
            artifacts_references_inputs = [_ensure_reference(item) for item in outputs if item is not None]
            self.add_outputs(artifacts_references_inputs)
        run_output = self._client.get_run(self.id)
        run_output["status"] = status
        run_output["endDate"] = datetime.utcnow()
        run_output["reason"] = reason
        rule_output = StopRunInput(run_output)
        if self._integration_client is not None:
            try:
                self._integration_client.before_stop(self)
            except Exception as e:
                logging.warning(e)
        output = self._client.stop_run(rule_output)
        self._status = output.jobRun.status
        self._duration = output.jobRun.duration
        self._startDate = output.jobRun.start_date
        self._endDate = output.jobRun.end_date
        if self._integration_client is not None:
            try:
                self._integration_client.after_stop(self)
            except Exception as e:
                logging.warning(e)

    def complete(
        self, outputs: Optional[List[Union[ArtifactReference, CodeVersion, ModelVersion, DatasetVersion]]] = None
    ) -> None:
        """
        Completes the run.

        :param outputs: The artifacts to be added as outputs to the run

        :return: None
        """
        self.end_run(outputs=outputs, status=RunStatus.COMPLETED)

    def fail(
        self,
        outputs: Optional[List[Union[ArtifactReference, CodeVersion, ModelVersion, DatasetVersion]]] = None,
        reason: Optional[str] = None,
    ) -> None:
        """
        Fails the run.

        :param outputs: The artifacts to be added as outputs to the run
        :param reason: Reason note for failing the run

        :return: None
        """
        self.end_run(outputs=outputs, status=RunStatus.FAILED, reason=reason)

    def abort(
        self,
        outputs: Optional[List[Union[ArtifactReference, CodeVersion, ModelVersion, DatasetVersion]]] = None,
        reason: Optional[str] = None,
    ) -> None:
        """
        Aborts the run.

        The end date of the Run will be set at the time this method is called.

        :param outputs: The artifacts to be added as outputs to the run
        :param reason: Reason note for aborting the run
        :return: None
        """
        self.end_run(outputs=outputs, status=RunStatus.ABORTED, reason=reason)

    def clean(self):
        """
        Cleans and rests to None all the data of the run.

        """
        self._outputs = None
        self._inputs = None
        self._endDate = None
        self._duration = None
        self._status = RunStatus.SCHEDULED

    def _capture_run_info(self, stage: Stage):
        inputs = self.inputs
        outputs = self.outputs
        logging.getLogger("Stage").propagate = False
        inputs_documentation(inputs, stage, self._client)
        # Run
        stage.add_block(divider=True)
        stage._add_header("The Original Run:")
        if self.description:
            stage.add_block(text=self.description)
        stage.add_block(text=" ")
        stage.add_block(run=self.id)
        stage.add_block(divider=True)

        outputs_documentation(outputs, stage, self._client)
        logging.getLogger("Stage").propagate = True



from __future__ import annotations

import urllib
from typing import Optional
from typing import TYPE_CHECKING
from urllib.parse import urlencode

from ._auth import Auth
from ._http import HttpError
from .http_error_handlers import HttpErrorHandler
from .json import DatasetOutput, DatasetInput, Page, PagedResponse
from .project import ProjectApi
from .reference import InvalidReferenceError, MissingReferenceError

if TYPE_CHECKING:
    from vectice import Reference


class DatasetApi:
    def __init__(self, auth: Auth):
        self._auth = auth
        self._httpErrorhandler = HttpErrorHandler()

    def get_dataset(
        self, dataset: Reference, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ) -> DatasetOutput:
        if not isinstance(dataset, int) and not isinstance(dataset, str):
            raise InvalidReferenceError("dataset", dataset)
        if isinstance(dataset, int):
            parent_project = None
            url = f"/metadata/dataset/{dataset}"
        else:
            if project is None:
                raise MissingReferenceError("dataset", "project")
            parent_project = ProjectApi(self._auth).get_project(project, workspace)
            url = f"/metadata/project/{parent_project.id}/dataset/name/{urllib.parse.quote(dataset)}"
        try:
            response = self._auth._get(url)
            result = DatasetOutput(**response)
            if parent_project is not None:
                result.project = parent_project
            return result
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "dataset", dataset)
        except IndexError:
            raise ValueError("The dataset is invalid. Please check the entered value.")

    def delete_dataset(
        self, dataset: Reference, project: Optional[Reference] = None, workspace: Optional[Reference] = None
    ):
        try:
            if isinstance(dataset, str):
                dataset_object = self.get_dataset(dataset, project, workspace)
                self._auth._delete(f"/metadata/dataset/{dataset_object.id}")
            else:
                self._auth._delete(f"/metadata/dataset/{dataset}")
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "dataset", dataset)

    def list_datasets(
        self,
        project: Reference,
        workspace: Optional[Reference] = None,
        search: Optional[str] = None,
        page_index=Page.index,
        page_size=Page.size,
    ) -> PagedResponse[DatasetOutput]:
        if isinstance(project, int):
            base_url = f"/metadata/project/{project}/dataset"
        elif isinstance(project, str):
            parent_project = ProjectApi(self._auth).get_project(project, workspace)
            base_url = f"/metadata/project/{parent_project.id}/dataset"
        else:
            raise InvalidReferenceError("project", project)
        try:
            queries = {"index": page_index, "size": page_size}
            if search:
                queries["search"] = search
            response = self._auth._get(f"{base_url}?{urlencode(queries)}")
            result = PagedResponse(
                item_cls=DatasetOutput, total=response["total"], page=response["page"], items=response["items"]
            )
            return result
        except HttpError as e:
            raise self._httpErrorhandler.handleGetHttpError(e, "project", project)

    def create_dataset(
        self, data: DatasetInput, project: Reference, workspace: Optional[Reference] = None
    ) -> DatasetOutput:
        if data.name is None:
            raise ValueError('"name" must be provided in dataset.')
        if isinstance(project, int):
            url = f"/metadata/project/{project}/dataset"
        elif isinstance(project, str):
            parent_project = ProjectApi(self._auth).get_project(project, workspace)
            url = f"/metadata/project/{parent_project.id}/dataset"
        else:
            raise InvalidReferenceError("project", project)
        try:
            response = self._auth._post(url, data)
            return DatasetOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePostHttpError(e, "dataset")

    def update_dataset(
        self,
        data: DatasetOutput,
        dataset: Reference,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> DatasetOutput:
        if data.name is None:
            raise ValueError('"name" must be provided in dataset.')
        if isinstance(dataset, int):
            url = f"/metadata/dataset/{dataset}"
        elif isinstance(dataset, str):
            dataset_object = self.get_dataset(dataset, project, workspace)
            url = f"/metadata/dataset/{dataset_object.id}"
        else:
            raise InvalidReferenceError("dataset", dataset)
        try:
            response = self._auth._put(url, data)
            return DatasetOutput(**response)
        except HttpError as e:
            raise self._httpErrorhandler.handlePutHttpError(e, "dataset", dataset)



from __future__ import annotations

import inspect
from dataclasses import dataclass, asdict
from datetime import datetime, timezone
from typing import Optional, List, Dict, Union, Any


class MetricInput:
    def __init__(self, key: str, value: Union[float, int], timestamp: Optional[Union[datetime, str]] = None):
        self.key: str = key
        """"""
        self.value: Union[float, int] = value
        """"""
        if timestamp is None:
            self.timestamp: str = datetime.now(timezone.utc).isoformat()
        else:
            self.timestamp = timestamp.isoformat() if isinstance(timestamp, datetime) else timestamp


@dataclass
class MetricOutput:
    key: str
    value: float
    timestamp: datetime
    name: Optional[str] = None
    id: Optional[int] = None

    @classmethod
    def from_dict(cls, metrics):
        return cls(**{k: v for k, v in metrics.items() if k in inspect.signature(cls).parameters})

    def as_dict(self):
        return asdict(self)


def create_metrics_input(metrics: Dict[str, float]) -> List[MetricInput]:
    if len(set(metrics)) < len(metrics):
        raise ValueError("You can not use the same key value pair more than once.")
    props: List[MetricInput] = []
    for key, value in metrics.items():
        _check_empty_property(key, value)
        props.append(MetricInput(key, value))
    return props


def _check_empty_property(key: str, value: Any):
    if key.strip() == "":
        raise ValueError("Property keys and values can't be empty.")



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from vectice.models import Run


class AbstractIntegration(ABC):
    @abstractmethod
    def lib_name(self) -> str:
        pass

    @abstractmethod
    def before_start(self, vectice_run: Run) -> None:
        """ """
        pass

    @abstractmethod
    def after_start(self, vectice_run: Run) -> None:
        """ """
        pass

    @abstractmethod
    def before_stop(self, vectice_run: Run) -> None:
        """ """
        pass

    @abstractmethod
    def after_stop(self, vectice_run: Run) -> None:
        """ """
        pass



from __future__ import annotations

import logging
from typing import Optional, Tuple, Any, BinaryIO, TYPE_CHECKING, Sequence, cast

from vectice.api._auth import Auth
from vectice.api.code_version import CodeVersionApi
from vectice.api.dataset_version import DatasetVersionApi
from vectice.api.model_version import ModelVersionApi
from vectice.api.project import ProjectApi
from vectice.api.run import RunApi
from ._http import HttpError
from .http_error_handlers import HttpErrorHandler
from .json import PagedResponse, AttachmentOutput
from .reference import InvalidReferenceError, MissingReferenceError

if TYPE_CHECKING:
    from vectice import Reference


class AttachmentApi:
    def __init__(self, auth: Auth):
        self._auth = auth
        self._httpErrorhandler = HttpErrorHandler()
        self._logger = logging.getLogger(self.__class__.__name__)

    def _generate_artifact_url_and_id(
        self,
        _type: str,
        artifact_version: Optional[Reference] = None,
        artifact: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> Tuple[str, Optional[str]]:
        url = None
        object_name = None
        try:
            if _type == "datasetversion":
                if isinstance(artifact_version, int):
                    dataset_version_object = DatasetVersionApi(self._auth).get_dataset_version(
                        version=artifact_version, dataset=artifact, project=project, workspace=workspace
                    )
                elif isinstance(artifact_version, str) and project:
                    dataset_version_object = DatasetVersionApi(self._auth).get_dataset_version(
                        version=artifact_version, dataset=artifact, project=project, workspace=workspace
                    )
                else:
                    raise MissingReferenceError("datasetversion", "project")
                url = f"/metadata/project/{dataset_version_object.dataset.project.id}/entityfiles/{_type}/{dataset_version_object.id}"
                object_name = "DatasetVersion with id: " + str(dataset_version_object.id)
            elif _type == "modelversion":
                if isinstance(artifact_version, int):
                    model_version_object = ModelVersionApi(self._auth).get_model_version(
                        version=artifact_version, model=artifact, project=project, workspace=workspace
                    )
                elif isinstance(artifact_version, str) and project:
                    model_version_object = ModelVersionApi(self._auth).get_model_version(
                        version=artifact_version, model=artifact, project=project, workspace=workspace
                    )
                else:
                    raise MissingReferenceError("modelversion", "project")
                url = f"/metadata/project/{model_version_object.model.project.id}/entityfiles/{_type}/{model_version_object.id}"
                object_name = "ModelVersion with id: " + str(model_version_object.id)
            elif _type == "codeversion":
                if isinstance(artifact_version, int):
                    code_version_object = CodeVersionApi(self._auth).get_code_version(
                        version=artifact_version, project=project, workspace=workspace
                    )
                elif isinstance(artifact_version, str) and project:
                    code_version_object = CodeVersionApi(self._auth).get_code_version(
                        version=artifact_version, project=project, workspace=workspace
                    )
                else:
                    raise MissingReferenceError("codeversion", "project")
                url = f"/metadata/project/{code_version_object.code.project.id}/entityfiles/{_type}/{code_version_object.id}"
                object_name = "CodeVersion with id: " + str(code_version_object.id)
            elif _type == "run":
                if isinstance(artifact_version, int):
                    run_object = RunApi(self._auth).get_run(
                        run=artifact_version, job=artifact, project=project, workspace=workspace
                    )
                elif isinstance(artifact_version, str) and project:
                    run_object = RunApi(self._auth).get_run(
                        run=artifact_version, job=artifact, project=project, workspace=workspace
                    )
                else:
                    raise MissingReferenceError("run", "project")
                url = f"/metadata/project/{run_object.job.project.id}/entityfiles/{_type}/{run_object.id}"
                object_name = "Run with id: " + str(run_object.id)
            elif _type == "project":
                if project:
                    project_object = ProjectApi(self._auth).get_project(project, workspace)
                else:
                    raise MissingReferenceError("project", "attachment")
                url = f"/metadata/project/{project_object.id}/entityfiles/{_type}/{project_object.id}"
            if url is None:
                raise RuntimeError("url cannot be none")
            return url, object_name
        except HttpError as e:
            if artifact_version:
                reference = artifact_version
            elif artifact:
                reference = artifact
            else:
                raise ValueError("No reference to artifact or artifact version provided.")
            raise self._httpErrorhandler.handleGetHttpError(e, _type, reference)

    def get_attachment(
        self,
        _type: str,
        file_id: int,
        artifact_version: Optional[Reference] = None,
        artifact: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> BinaryIO:
        try:
            url, object_id = self._generate_artifact_url_and_id(_type, artifact_version, artifact, workspace, project)
            response = self._auth._get_attachment(url + f"/{file_id}")
            self._logger.info(f"Attachment with id: {file_id} successfully retrieved from {object_id}.")
            return cast(BinaryIO, response.raw)
        except HttpError as e:
            if artifact_version:
                reference = artifact_version
            elif artifact:
                reference = artifact
            else:
                raise ValueError("No reference to artifact or artifact version provided.")
            raise self._httpErrorhandler.handleGetHttpError(e, _type, reference)

    def post_attachment(
        self,
        _type: str,
        artifact_version: Optional[Reference] = None,
        files: Optional[Sequence[Tuple[str, Tuple[Any, BinaryIO]]]] = None,
        artifact: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
        file_id: Optional[int] = None,
    ) -> None:
        try:
            url, object_id = self._generate_artifact_url_and_id(_type, artifact_version, artifact, workspace, project)
            if file_id:
                url = url + f"/{file_id}"
                self._logger.info(f"Attachment with id: {file_id} successfully attached to {object_id}.")
            if files and len(files) == 1:
                self._auth._post_attachments(url, files)
                self._logger.info(f"Attachment with name: {files[0][1][0]} successfully attached to {object_id}.")
            elif files and len(files) > 1:
                for file in files:
                    self._auth._post_attachments(url, [file])
                self._logger.info(
                    f"Attachments with names: {[f[1][0] for f in files]} successfully attached to {object_id}."
                )
        except HttpError as e:
            if artifact_version:
                reference = artifact_version
            elif artifact:
                reference = artifact
            elif _type == "project" and project:
                reference = project
            else:
                raise ValueError("No reference to artifact or artifact version provided.")
            raise self._httpErrorhandler.handleGetHttpError(e, _type, reference)

    def delete_attachment(
        self,
        _type: str,
        artifact_version: Reference,
        file_id: int,
        artifact: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ):
        try:
            url, object_id = self._generate_artifact_url_and_id(_type, artifact_version, artifact, workspace, project)
            response = self._auth._delete_attachment(url + f"/{file_id}")
            self._logger.info(f"Attachment with id: {file_id} successfully deleted from {object_id}.")
            return response
        except HttpError as e:
            if artifact_version:
                reference = artifact_version
            elif artifact:
                reference = artifact
            else:
                raise ValueError("No reference to artifact or artifact version provided.")
            raise self._httpErrorhandler.handleGetHttpError(e, _type, reference)

    def list_attachments(
        self,
        _type: str,
        artifact_version: Optional[Reference] = None,
        artifact: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ) -> PagedResponse[AttachmentOutput]:
        url = None
        try:
            url, _ = self._generate_artifact_url_and_id(_type, artifact_version, artifact, workspace, project)
        except HttpError as e:
            if artifact_version:
                reference = artifact_version
            else:
                raise ValueError("No reference to artifact version provided.")
            self._httpErrorhandler.handleGetHttpError(e, _type, reference)
        if url is None:
            raise InvalidReferenceError("artifact version", artifact)
        attachments = self._auth._list_attachments(url)
        return PagedResponse(
            item_cls=AttachmentOutput,
            total=len(attachments),
            page={},
            items=attachments,
        )

    def update_attachments(
        self,
        _type: str,
        files: Sequence[Tuple[str, Tuple[Any, BinaryIO]]],
        artifact_version: Optional[Reference] = None,
        artifact: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        project: Optional[Reference] = None,
    ):
        try:
            url, object_id = self._generate_artifact_url_and_id(_type, artifact_version, artifact, workspace, project)
            attachments = {
                attach.fileName: attach.fileId
                for attach in self.list_attachments(_type, artifact_version, artifact, workspace, project).list
            }
            for file in files:
                file_name = file[1][0]
                file_id = attachments.get(file_name)
                if file_id:
                    self._auth._put_attachments(url + f"/{file_id}", [file])
            self._logger.info(
                f"Attachments with names: {[f[1][0] for f in files]} successfully updated in {object_id}."
            )
        except HttpError as e:
            if artifact_version:
                reference = artifact_version
            elif artifact:
                reference = artifact
            else:
                raise ValueError("No reference to artifact or artifact version provided.")
            self._httpErrorhandler.handleGetHttpError(e, _type, reference)



from __future__ import annotations

from typing import Optional, TYPE_CHECKING

from vectice.api import Client

if TYPE_CHECKING:
    from vectice.models import Project


class Code:
    """
    Describe a code


    """

    def __init__(
        self,
        name: str,
        id: int,
        project: Project,
        description: Optional[str] = None,
        uri: Optional[str] = None,
        git_version: Optional[int] = None,
    ):
        """
        :param id: the project identifier
        :param project: the project the code belong to
        :param name: the name of the code
        :param description: a quick  description of the code
        :param uri:
        """
        self._id = id
        self._name = name
        self._project: Project = project
        self._description = description
        self._uri = uri
        self._git_version = git_version
        self._client: Client = project._client

    @property
    def id(self) -> Optional[int]:
        """
        The code identifier
        :return:
        """
        return self._id

    @id.setter
    def id(self, code_id: int):
        self._id = code_id

    @property
    def project(self) -> Optional[Project]:
        """
        The project this code belong to
        """
        return self._project

    @property
    def name(self) -> str:
        """
        Name of the code
        """
        return self._name

    @property
    def description(self) -> Optional[str]:
        """
        Quick description of the code
        """
        return self._description

    @description.setter
    def description(self, description: str):
        self._description = description

    @property
    def uri(self) -> Optional[str]:
        return self._uri

    @uri.setter
    def uri(self, uri: str):
        self._uri = uri

    @property
    def git_version(self) -> Optional[int]:
        return self._git_version

    @git_version.setter
    def git_version(self, git_version: int):
        self._git_version = git_version



from __future__ import annotations

import logging
from datetime import datetime
from typing import Optional, List

import mlflow
from mlflow.entities import Experiment as MlflowExperiment, Run as MlflowRun
from mlflow.tracking import MlflowClient

from vectice.api import Client
from vectice.api.json import ModelInput
from vectice.api.json import PropertyInput
from vectice.api.json import RunStatus
from vectice.integrations import AbstractIntegration
from vectice.models.artifact_reference import ArtifactReference
from vectice.models.run import Run

VECTICE_RUN_TAG = "vectice.run.id"
_STATUSES_MAP = {
    "RUNNING": RunStatus.STARTED,
    "SCHEDULED": RunStatus.SCHEDULED,
    "FINISHED": RunStatus.COMPLETED,
    "FAILED": RunStatus.FAILED,
    "KILLED": RunStatus.ABORTED,
}

_STATUSES_VECTICE_TO_MLFLOW = {
    "STARTED": "RUNNING",
    "SCHEDULED": "SCHEDULED",
    "COMPLETED": "FINISHED",
    "FAILED": "FAILED",
    "ABORTED": "KILLED",
}

UNWANTED_TAGS = ["mlflow.log-model.history"]


class MLflowIntegration(AbstractIntegration):
    def __init__(self, mlflow_client: MlflowClient, vectice_client: Client, auto_log: Optional[bool] = False):

        self._auto_log = auto_log
        self._mlflow_client = mlflow_client
        self._vectice_client = vectice_client
        self._auto_log = auto_log
        self._logger = logging.getLogger(self.__class__.__name__)
        self._experiment_name: Optional[str] = None
        self._mlflow_run = None
        self._output_artifacts: List[ArtifactReference] = []

    @property
    def experiment_name(self):
        return self._experiment_name

    @property
    def lib_name(self) -> str:
        return "mlflow"

    def before_start(self, vectice_run: Run):
        mlflow.set_experiment(vectice_run.job.name)
        self._experiment_name = vectice_run.job.name

    def after_start(self, vectice_run: Run) -> None:
        """
        Updated the existing job.
        """
        ml_run = mlflow.start_run()
        self._mlflow_run = ml_run

    def before_stop(self, vectice_run: Run):
        """
        Might be useful.
        """
        pass

    def after_stop(self, vectice_run: Run):
        """
        Clean up.
        """
        if mlflow.active_run() is None:
            self._logger.debug("No active run found.")
            return None
        status = _STATUSES_VECTICE_TO_MLFLOW[vectice_run.status.name]
        mlflow.end_run(status=status)
        self._output_artifacts = self._extract_from_mlflow(self._mlflow_run, vectice_run)
        if self._mlflow_run is not None:
            self._mlflow_client.set_tag(self._mlflow_run.info.run_id, VECTICE_RUN_TAG, vectice_run.id)
        elif self._mlflow_run is None:
            logging.warning("There is no MLflow run.")
        self._experiment_name = vectice_run.job.name
        vectice_run.add_outputs(self._output_artifacts)

    def _extract_from_mlflow(
        self, mlflow_run: MlflowRun, vectice_run: Run, experiment: MlflowExperiment = None
    ) -> List[ArtifactReference]:
        ml_exp = self._mlflow_client.get_experiment(mlflow_run.info.experiment_id)
        if vectice_run is not None:
            output_artifacts = self._extract_outputs(mlflow_run, ml_exp, vectice_run)
        else:
            output_artifacts = []
        self._extract_run(mlflow_run, vectice_run)
        return output_artifacts

    def _extract_run(self, mlflow_run: MlflowRun, vectice_run: Run):
        """
        Updated run
        """
        from vectice.api.json.run import RunOutput

        tags = []
        run_name = None
        for tag in mlflow_run.data.tags.items():
            if tag[0] not in UNWANTED_TAGS:
                tags.append(PropertyInput(tag[0], tag[1]))
        if vectice_run:
            run_name = vectice_run.name
        self._vectice_client.create_run_properties(vectice_run.id, tags)
        if vectice_run is not None:
            run_input = RunOutput(
                id=vectice_run.id,
                start_date=datetime.fromtimestamp(mlflow_run.info.start_time / 1000),
                end_date=datetime.fromtimestamp(mlflow_run.info.end_time / 1000) if mlflow_run.info.end_time else None,
                name=run_name if run_name else None,
            )
            self._vectice_client.update_run(run_input)

    def _extract_outputs(
        self, mlflow_run: MlflowRun, experiment: MlflowExperiment, run: Run
    ) -> List[ArtifactReference]:
        from vectice.api.json import ModelType
        from vectice.models.model import Model
        from vectice.api.reference import BadReferenceError

        updated_run = self._mlflow_client.get_run(mlflow_run.info.run_id)

        parameters = updated_run.data.params
        metrics = updated_run.data.metrics
        parameters[
            "mlflow_url"
        ] = f"{mlflow.get_tracking_uri()}/#/experiments/{mlflow_run.info.experiment_id}/runs/{mlflow_run.info.run_id}"
        try:
            model_output = self._vectice_client.get_model(experiment.name)
        except BadReferenceError:
            model_output = None
        if model_output is None:
            model_input = ModelInput(name=experiment.name, type=ModelType.OTHER.name)
            model_output = self._vectice_client.create_model(data=model_input, project=run.job.project.id)
        model = Model(
            model_output.name, model_output.id, run.job.project, model_output.description, ModelType[model_output.type]
        )
        model_version = model.create_model_version(
            experiment.name + "-" + mlflow_run.info.run_id,
            algorithm_name=mlflow_run.data.tags.get("estimator_name"),
            metrics=metrics,
            hyper_parameters=parameters,
            runId=run.id,
        )
        result = ArtifactReference(model=model.id, version_id=model_version.id)
        return [result]



from __future__ import annotations

from typing import Optional, List, Union, BinaryIO, TYPE_CHECKING, Tuple

from .attachment_container import AttachmentContainer
from .git_version import GitVersion

if TYPE_CHECKING:
    from vectice import Reference
    from vectice.models.project import Project


# @dataclass
# class __CodeData:
#     gitVersion: GitVersion
#     """
#     git information structure extracted automatically by the SDK.
#     """


class CodeVersion(AttachmentContainer):
    """
    Code version class
    """

    def __init__(
        self,
        project: Project,
        id: int,
        name: str,
        version_number: Optional[int] = None,  # for rule API
        code_id: Optional[int] = None,
        description: Optional[str] = None,
        uri: Optional[str] = None,
        is_starred: Optional[bool] = False,
        attachments: Optional[Union[str, List[str]]] = None,
        git_version_id: Optional[int] = None,
        git_version: Optional[GitVersion] = None,
        version: Optional[Reference] = None,
    ):
        super().__init__(name, id, project._client, "CodeVersion")
        self._project = project
        self._description = description
        self._uri = uri
        self._isStarred = is_starred
        self._attachments: Optional[
            Union[Tuple[str, Tuple[str, BinaryIO]], List[Tuple[str, Tuple[str, BinaryIO]]]]
        ] = None
        self._version = version
        self._git_version_id = git_version_id
        self._git_version = git_version
        self._code_id = code_id
        self._version_number = version_number

        if git_version is None:
            pass
        if attachments:
            self.add_attachments(attachments)

    def __repr__(self):
        return f"CodeVersion(id={self.id}, name={self.name}, description={self.description}, uri={self.uri}, gitVersion={self.git_version}, project={self.project})"

    @property
    def project(self) -> Project:
        return self._project

    @property
    def id(self) -> int:
        """
        The code identifier
        :return: int
        """
        if self._id is None:
            raise RuntimeError("can not use id as the code version is not saved")
        return self._id

    @property
    def code_id(self) -> Optional[int]:
        """
        The code identifier
        :return: Optional[int]
        """
        return self._code_id

    @property
    def name(self) -> Optional[str]:
        """
        Name of the code
        :return: Optional[str]
        """
        return self._name

    @property
    def version(self) -> Optional[Reference]:
        """
        The version of the code
        :return: Optional[Reference]
        """
        return self._version

    @property
    def description(self) -> Optional[str]:
        """
        Quick description of the code
        """
        return self._description

    @description.setter
    def description(self, description: str):
        self._description = description

    @property
    def uri(self) -> Optional[str]:
        """
        The uri to the code
        :return: Optional[str]
        """
        return self._uri

    @uri.setter
    def uri(self, uri: str):
        self._uri = uri

    @property
    def is_starred(self) -> Optional[bool]:
        """
        A boolean to star the code
        :return: Optional[bool]
        """
        return self._isStarred

    @is_starred.setter
    def is_starred(self, is_starred: bool):
        self._isStarred = is_starred

    @property
    def version_number(self) -> Optional[int]:
        """
        The version number the code
        :return: Optional[int]
        """
        return self._version_number

    @property
    def attachments(self) -> Optional[Union[Tuple[str, Tuple[str, BinaryIO]], List[Tuple[str, Tuple[str, BinaryIO]]]]]:
        """
        The attachments linked to the code
        """
        return self._attachments

    @attachments.setter
    def attachments(self, attachments: Union[Tuple[str, Tuple[str, BinaryIO]], List[Tuple[str, Tuple[str, BinaryIO]]]]):
        self._attachments = attachments

    @property
    def git_version_id(self) -> Optional[int]:
        """
        The git version id of the code
        :return: Optional[int]
        """
        return self._git_version_id

    @property
    def git_version(self) -> Optional[GitVersion]:
        """
        The GitVersion of the code
        :return: Optional[GitVersion]
        """
        return self._git_version



from __future__ import annotations

import urllib
from typing import Optional, TYPE_CHECKING, Dict, Any
from urllib.parse import urlencode

from vectice.api._auth import Auth
from ._http import HttpError
from .http_error_handlers import HttpErrorHandler
from .json import ConnectionOutput, Page, PagedResponse, ConnectionInput
from .reference import InvalidReferenceError
from .workspace import WorkspaceApi

if TYPE_CHECKING:
    from vectice import Reference


class ConnectionApi:
    def __init__(self, auth: Auth):
        self._auth = auth
        self._httpErrorHandler = HttpErrorHandler()

    def _create_base_url(self, workspace: Reference) -> str:
        if isinstance(workspace, str):
            parent_workspace = WorkspaceApi(self._auth).get_workspace(workspace)
            return f"/metadata/workspace/{parent_workspace.id}"
        elif isinstance(workspace, int):
            return f"/metadata/workspace/{workspace}"
        else:
            raise InvalidReferenceError("workspace", workspace)

    def get_connection(self, connection: Reference, workspace: Optional[Reference] = None) -> ConnectionOutput:
        if isinstance(connection, int):
            url = f"/metadata/connection/{connection}"
        elif isinstance(connection, str):
            if isinstance(workspace, str):
                parent_workspace = WorkspaceApi(self._auth).get_workspace(workspace)
                url = f"/metadata/workspace/{parent_workspace.id}/connection/name/{urllib.parse.quote(connection)}"
            elif isinstance(workspace, int):
                url = f"/metadata/workspace/{workspace}/connection/name/{urllib.parse.quote(connection)}"
            else:
                raise InvalidReferenceError("workspace", workspace)
        else:
            raise InvalidReferenceError("connection", connection)
        try:
            response = self._auth._get(url)
        except HttpError as e:
            raise self._httpErrorHandler.handleGetHttpError(e, "connection", connection)
        return ConnectionOutput(**response)

    def list_connections(
        self,
        workspace: Reference,
        connection_type: Optional[str] = None,
        search: Optional[str] = None,
        page_index: int = Page.index,
        page_size: int = Page.size,
    ) -> PagedResponse[ConnectionOutput]:
        url = self._create_base_url(workspace)
        queries: Dict[str, Any] = {"index": page_index, "size": page_size}
        if search:
            queries["search"] = search
        if connection_type:
            queries["type"] = connection_type
        try:
            response = self._auth._get(url + "/connection" + "?" + urlencode(queries))
            return PagedResponse(
                item_cls=ConnectionOutput,
                total=response["total"],
                page=response["page"],
                items=response["items"],
            )
        except HttpError as e:
            raise self._httpErrorHandler.handleGetHttpError(e, "workspace", workspace)

    def delete_connection(self, connection: Reference, workspace: Reference) -> None:
        if not isinstance(connection, int) and not isinstance(connection, str):
            raise ValueError("Please provide a valid connection reference.")
        connection_output = self.get_connection(connection, workspace)
        try:
            self._auth._delete(f"/metadata/connection/{connection_output.id}")
        except HttpError as e:
            raise self._httpErrorHandler.handleDeleteHttpError(e, "connection", connection)

    def create_connection(self, data: ConnectionInput, workspace: Reference) -> ConnectionOutput:
        parent_workspace = WorkspaceApi(self._auth).get_workspace(workspace)
        url = f"/metadata/workspace/{parent_workspace.id}/connection"
        response = self._auth._post(url, data)
        return ConnectionOutput(**response)

    def update_connection(self, data: ConnectionInput, connection: Reference, workspace: Reference) -> ConnectionOutput:
        if not isinstance(connection, int) and not isinstance(connection, str):
            raise ValueError("Please provide a valid connection reference.")
        parent_workspace = WorkspaceApi(self._auth).get_workspace(workspace)
        if isinstance(connection, int):
            response = self._auth._put(f"/metadata/workspace/{parent_workspace.id}/connection/{connection}", data)
            return ConnectionOutput(**response)
        else:
            connection_output = self.get_connection(connection, parent_workspace.id)
            response = self._auth._put(
                f"/metadata/workspace/{parent_workspace.id}/connection/{connection_output.id}", data
            )
            return ConnectionOutput(**response)



from __future__ import annotations

import logging
from datetime import datetime
from typing import Optional, Dict, List, TYPE_CHECKING

from vectice.api import Reference, Client
from vectice.api.json import (
    ModelType,
    ModelVersionInput,
    ModelVersionStatus,
    Page,
    PropertyInput,
)
from vectice.api.json.metric import create_metrics_input
from vectice.api.json.property import create_properties_input
from .model_version import ModelVersion
from .property import create_properties

if TYPE_CHECKING:
    from vectice.models import Project


class Model:
    """
    Describes a model

    Noteworthy:
    1. id: Backend will return id if it doesn't exist so it should be optional. SDK doesn't assign id unless we retrieve or resuse a model

    """

    def __init__(
        self,
        name: str,
        id: int,
        project: Project,
        description: Optional[str] = None,
        type: Optional[ModelType] = None,
    ):
        """
        :param name: The name of the model
        :param id: The model id
        :param project: The project this model belongs to
        :param description: The quick description of the model
        :param type: Type of the model
        """
        self._id = id
        self._project = project
        self._name = name
        self._type = type
        self._description = description
        self._client: Client = project._client
        self._logger = logging.getLogger(self.__class__.__name__)

    def __repr__(self):
        return f"Model(name={self.name}, id={self.id}, description={self.description}, type={self._type})"

    @property
    def id(self) -> int:
        """
        Model identifier.
        :return: int
        """
        return self._id

    @property
    def project(self) -> Project:
        """
        The project this model belong to.
        :return: Project
        """
        return self._project

    @property
    def name(self) -> str:
        """
        Name of the model.
        :return: str
        """
        return self._name

    @property
    def description(self) -> Optional[str]:
        """
        Quick description of the model.
        :return: Optional[str]
        """
        return self._description

    @property
    def type(self) -> Optional[ModelType]:
        """
        Type of the model.
        :return: Optional[ModelType]
        """
        return self._type

    def __eq__(self, o: object) -> bool:
        return (
            isinstance(o, Model)
            and self.id == o.id
            and self.name == o.name
            and o.description == self.description
            and self.type == o.type
        )

    def create_model_version(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        status: ModelVersionStatus = ModelVersionStatus.EXPERIMENTATION,
        algorithm_name: Optional[str] = None,
        is_starred: Optional[bool] = False,
        metrics: Optional[Dict] = None,
        hyper_parameters: Optional[Dict] = None,
        attachments: Optional[List[str]] = None,
        runId: Optional[int] = None,
    ) -> ModelVersion:
        """
        Creates a new model version with all the existing params in Vectice.

        :param name: The name of the model version
        :param description: The description of the model
        :param status: The status of the model e.g 'OTHER'/'CLASSIFICATION'/'REGRESSION'
        :param algorithm_name: The algorithm the model version uses
        :param is_starred: Whether the model is starred or not
        :param metrics: The model version metrics
        :param hyper_parameters: The model version hyper parameters
        :param attachments: Attach file/s to the model version
        :param runId: Link the model version to a run

        :return: A ModelVersion
        """
        self._logger.propagate = True
        data = ModelVersionInput(
            name=name,
            description=description,
            algorithmName=algorithm_name,
            status=status.value,
            isStarred=is_starred,
        )
        if runId is not None:
            data["jobRunId"] = runId
        model_version_output = self._client.create_model_version(data, self._id)
        if hyper_parameters:
            self._client.create_model_version_properties(
                model_version_output.id, create_properties_input(hyper_parameters)
            )
        if metrics:
            self._client.create_model_version_metrics(model_version_output.id, create_metrics_input(metrics))

        model_version = ModelVersion(
            self,
            model_version_output.id,
            model_version_output.name,
            model_version_output.version_number,
            model_version_output.description,
            metrics,
            hyper_parameters,
            model_version_output.algorithm_name,
            model_version_output.status,
            model_version_output.is_starred,
            created_date=model_version_output.created_date if model_version_output.created_date else datetime.now(),
            updated_date=model_version_output.updated_date,
            deleted_date=model_version_output.deleted_date,
        )
        if attachments:
            model_version.add_attachments(attachments)
        self._logger.info(f"ModelVersion with id: {model_version.id} successfully created in Model {self.name}.")
        return model_version

    def update_model_version(
        self,
        version: Reference,
        name: Optional[str] = None,
        description: Optional[str] = None,
        status: Optional[str] = "EXPERIMENTATION",
        algorithm: Optional[str] = None,
        is_starred: Optional[bool] = False,
        metrics: Optional[Dict] = None,
        hyper_parameters: Optional[Dict] = None,
    ) -> ModelVersion:
        """
        Update a model version.
        Updates the specified model version with the new params passed to this function.
        :param version: The model version name or id
        :param name: The name of the model version
        :param description: The description of the model
        :param status: The status of the model e.g 'EXPERIMENTATION'/'STAGING'/'PRODUCTION'
        :param algorithm: The algorithm the model version uses
        :param is_starred: Whether the model is starred or not
        :param metrics: The model version metrics
        :param hyper_parameters: The model version hyper parameters
        :return: ModelVersion
        """
        self._logger.propagate = True
        model_version_input = ModelVersionInput(
            name=name,
            description=description,
            isStarred=is_starred,
            status=status,
            algorithmName=algorithm,
            metrics=metrics,
            hyper_parameters=hyper_parameters,
        )
        model_version_output = self._client.update_model_version(model_version_input, version, self._id)
        if hyper_parameters:
            new_parameters = create_properties(hyper_parameters)
            parameters = {
                prop.key: prop.id
                for prop in self._client.list_model_version_properties(model_version_output.id, self._id).list
            }
            for param in new_parameters:
                if parameters.get(param.key):
                    self._client.update_model_version_properties(
                        model_version_output.id, parameters[param.key], [PropertyInput(param.key, param.value)]
                    )
        if metrics:
            new_metrics = create_metrics_input(metrics)
            metrics_list = {
                metric.key: metric.id
                for metric in self._client.list_model_version_metrics(model_version_output.id, self._id).list
            }
            for metric in new_metrics:
                temp = metrics_list.get(metric.key)
                if temp is not None:
                    self._client.update_model_version_metrics(self.id, model_version_output.id, temp, [metric])
        self._logger.info(f"ModelVersion with id: {model_version_output.id} successfully updated in Model {self.name}.")
        return ModelVersion(
            self,
            model_version_output.id,
            model_version_output.name,
            model_version_output.version_number,
            model_version_output.description,
            metrics,
            hyper_parameters,
            model_version_output.algorithm_name,
            model_version_output.status,
            model_version_output.is_starred,
            model_version_output.user_declared_version,
            model_version_output.created_date if model_version_output.created_date else datetime.now(),
            model_version_output.updated_date,
            model_version_output.deleted_date,
        )

    def list_model_versions(
        self, search: Optional[str] = None, page_index: int = Page.index, page_size: int = Page.size
    ) -> List[ModelVersion]:
        """
        Lists all the versions created out of the specified model.

        :param search: The name to search
        :param page_index:
        :param page_size:

        :return: List of ModelVersions
        """

        def get_parameters(id: int):
            properties_output = self._client.list_model_version_properties(id, self._id)
            parameters = {item.key: item.value for item in properties_output.list}
            return parameters

        model_versions = self._client.list_model_versions(self.id, None, None, search, page_index, page_size)
        return [
            ModelVersion(
                self,
                model_version.id,
                model_version.name,
                model_version.version_number,
                model_version.description,
                model_version.metrics,
                get_parameters(model_version.id),
                model_version.algorithm_name,
                model_version.status,
                model_version.is_starred,
                model_version.user_declared_version,
                model_version.created_date if model_version.created_date else datetime.now(),
                model_version.updated_date,
                model_version.deleted_date,
            )
            for model_version in model_versions.list
        ]

    def list_model_versions_dataframe(
        self,
        search: Optional[str] = None,
        page_index: int = Page.index,
        page_size: int = Page.size,
    ) -> "pandas.DataFrame":  # type: ignore # noqa F821
        """
        Lists model versions in a pandas DataFrame and sorts by update date. Requires the pandas module to be installed,
        which can be done with ```pip install pandas``` or ```pip install vectice[pandas]```

        :param search: filter on name
        :param page_index: The page index. For example 1/20 would return page 1
        :param page_size: The number of versions on a page
        :return: pd.DataFrame
        """
        import pandas as pd  # type: ignore

        model_versions = self._client.list_model_versions(self.id, None, None, search, page_index, page_size)
        metric_dataframes = []
        property_dataframes = []
        for idx, model_version in enumerate(model_versions.list, start=0):
            model_metrics = self._client.list_model_version_metrics(model_version.id, 1, 1000).list
            temp_df_metrics = pd.DataFrame(
                data={metrics.key: [metrics.value] for metrics in model_metrics}, index=[idx]
            )
            metric_dataframes += [temp_df_metrics]

        for idx, model_version in enumerate(model_versions.list, start=0):
            models_properties = self._client.list_model_version_properties(model_version.id, 1, 1000).list
            temp_df_metrics = pd.DataFrame(
                data={metrics.key: [metrics.value] for metrics in models_properties}, index=[idx]
            )
            property_dataframes += [temp_df_metrics]

        metric_dataframes = pd.concat(metric_dataframes)
        property_dataframes = pd.concat(property_dataframes)
        merged_entities = pd.merge(metric_dataframes, property_dataframes, left_index=True, right_index=True)
        df_model = pd.DataFrame(
            data=model_versions.list,
            # "created_date", "name", "version_number", "status", "algorithm_name", "is_starred"
            columns=["createdDate", "name", "versionNumber", "status", "algorithmName", "isStarred"],
        )
        return pd.merge(df_model, merged_entities, left_index=True, right_index=True).sort_values(
            by=["versionNumber"], ascending=False
        )

    def get_model_version(
        self,
        version: Reference,
    ) -> ModelVersion:
        """
        Gets a specific version of a model.

        :param version: The model version name or id
        :return: A ModelVersion
        """
        self._logger.propagate = True
        model_version_output = self._client.get_model_version(version=version, model=self._id)
        properties_output = self._client.list_model_version_properties(model_version_output.id, self.id)
        parameters = {item.key: item.value for item in properties_output.list}
        metrics_output = self._client.list_model_version_metrics(model_version_output.id, self.id)
        metrics = {item.key: item.value for item in metrics_output.list}
        self._logger.info(
            f"ModelVersion with id: {model_version_output.id} successfully retrieved from Model {self.name}."
        )
        return ModelVersion(
            self,
            model_version_output.id,
            model_version_output.name,
            model_version_output.version_number,
            model_version_output.description,
            metrics,
            parameters,
            model_version_output.algorithm_name,
            model_version_output.status,
            model_version_output.is_starred,
            model_version_output.user_declared_version,
            model_version_output.created_date if model_version_output.created_date else datetime.now(),
            model_version_output.updated_date,
            model_version_output.deleted_date,
        )

    def delete_model_version(
        self,
        version: Reference,
    ) -> None:
        """
        Deletes the specified version of a model.

        :param version: The model version name or id
        :return: None
        """
        self._logger.propagate = True
        model_version = self._client.get_model_version(version, self._id)
        self._client.delete_model_version(version, self._id)
        self._logger.info(f"ModelVersion with id: {model_version.id} successfully deleted from Model {self.name}.")

    def delete_model_version_properties(self, version: Reference, properties: List[str]):
        """
        Deletes the specified properties of a model version.

        :param version: The model version name or id
        :param properties: List of property names to delete.
        :return: None
        """
        property_ids = {item.key: item.id for item in self._client.list_model_version_properties(version, self.id).list}
        for prop in properties:
            if property_ids.get(prop):
                self._client.delete_model_version_properties(version, property_ids[prop], self.id)
            else:
                logging.warning(f"Hyper Parameter {prop} was not be found.")

    def delete_model_version_metrics(self, version: Reference, metrics: List[str]):
        """
        Deletes the specified metrics of a model version.

        :param version: The model version name or id
        :param metrics: List of metric names to delete.
        :return: None
        """
        metric_ids = {item.key: item.id for item in self._client.list_model_version_metrics(version, self.id).list}
        for metric in metrics:
            temp = metric_ids.get(metric)
            if temp is not None:
                self._client.delete_model_version_metrics(version, temp, self.id)
            else:
                logging.warning(f"Metric {metric} was not be found.")



from enum import Enum


class ArtifactType(Enum):
    """
    Enumeration of supported artifact types.
    """

    MODEL = "MODEL"
    """
    """

    CODE = "CODE"
    """
    """

    DATASET = "DATASET"
    """
    """


class JobArtifactType(Enum):
    """
    Indicates whether the artifact is an input or an output of a run.
    """

    INPUT = "INPUT"
    """
    """
    OUTPUT = "OUTPUT"
    """
    """



from typing import Optional
from urllib.parse import urlencode

from ._auth import Auth
from .json import ArtifactInput, ArtifactOutput, PagedResponse, Page
from .reference import Reference
from .run import RunApi


class ArtifactApi:
    def __init__(self, auth: Auth):
        self._auth = auth

    def create_artifact(
        self,
        data: ArtifactInput,
        run: Reference,
        job: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
    ) -> ArtifactOutput:
        if data.artifactType is None:
            raise ValueError('"artifactType" must be provided in artifact.')
        if data.jobArtifactType is None:
            raise ValueError('"jobArtifactType" must be provided in artifact.')
        parent_run = RunApi(self._auth).get_run(run, job, project, workspace)
        url = f"/metadata/project/{parent_run.job.project.id}/job/{parent_run.job.id}/run/{parent_run.id}/artifact"
        response = self._auth._post(url, data.__dict__)
        return ArtifactOutput(**response)

    def list_artifacts(
        self,
        run: Reference,
        job: Optional[Reference] = None,
        project: Optional[Reference] = None,
        workspace: Optional[Reference] = None,
        page_index: int = Page.index,
        page_size: int = Page.size,
    ) -> PagedResponse[ArtifactOutput]:
        parent_run = RunApi(self._auth).get_run(run, job, project, workspace)
        url = f"/metadata/project/{parent_run.job.project.id}/job/{parent_run.job.id}/run/{parent_run.id}/artifact"
        queries = {"index": page_index, "size": page_size}
        artifacts = self._auth._get(url + "?" + urlencode(queries))
        return PagedResponse(
            item_cls=ArtifactOutput,
            total=int(artifacts["total"]),
            page=artifacts["page"],
            items=artifacts["items"],
        )

    def update_artifact(self, artifact: ArtifactOutput) -> ArtifactOutput:
        pass
        # url = f"/metadata/project/{parent_run.job.project.id}/job/{parent_run.job.id}/run/{parent_run.id}/artifact/{artifact.id}"
        # return ArtifactOutput(self._auth._put(self.api_base_path + "/" + str(artifact_id), artifact))



from __future__ import annotations

import logging
from typing import List, Optional, Dict, TYPE_CHECKING, Any

from vectice.api.json import ProjectInput, ConnectionInput, ConnectionType
from .connection import Connection
from .project import Project

if TYPE_CHECKING:
    from vectice import Reference
    from vectice.api import Client
    from vectice.models.integration import AbstractIntegration


class Workspace:
    """
    Workspace is a space where you can found :py:class:`Project` and :py:class:`Connection` for a team.

    In a workspace, you can manage (create, update, delete, get and list ) projects.

    You can also lists and get connections but you can not create or delete them.

    """

    def __init__(self, id: int, name: str, description: Optional[str] = None):
        """
        :param id: the workspace identifier
        :param name: the name of the workspace
        :param description: the description of the workspace
        """
        self._id = id
        self._name = name
        self._description = description
        self._client: Client
        self._logger = logging.getLogger(self.__class__.__name__)

    def __post_init__(self, client: Client, integration_client: Optional[AbstractIntegration]):
        self._client = client
        self._integration_client = integration_client

    def __repr__(self):
        return f"Workspace(name={self.name}, id={self.id}, description={self.description})"

    @property
    def id(self) -> int:
        """
        The workspace identifier.
        :return: int
        """
        return self._id

    @property
    def name(self) -> str:
        """
        The name of the workspace.
        :return: str
        """
        return self._name

    @property
    def description(self) -> Optional[str]:
        """
        The description of the workspace.
        :return: Optional[str]
        """
        return self._description

    def create_project(self, name: str, description: Optional[str] = None) -> Project:
        """
        Creates a new project in the workspace with the given attributes.
        the name is required and must be unique in the workspace.

        By default, the project is public, change `is_public` to `False`
        to make it private.
        the status can be `New`, `InProgress` or `Deployed`

        :param name: The name of the project to be created.
        :param description: The description of the project

        :return: The newly created project
        """
        data = ProjectInput(name=name, description=description)
        output = self._client.create_project(data, self.id)
        self._logger.info(f"Project with id: {output.id} successfully created.")
        return Project(output.id, self, output.name, output.description)

    def update_project(
        self,
        project: Reference,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Project:
        """
        Updates a project.

        :param project: The project name or id to update.
        :param name: The name of the project to be.
        :param description: The description of the project

        :return: The updated project
        """
        data = ProjectInput(name=name, description=description)
        item = self._client.get_project(project, self.id)
        output = self._client.update_project(data, item.id, self.id)
        self._logger.info(f"Project with id: {item.id} successfully updated.")
        return Project(output.id, self, output.name, output.description)

    def get_project(self, project: Reference) -> Project:
        """
        Gets a project.

        Gets a project instance with the specified project name or id in the current workspace.

        :param project: The project name or id to get.

        :return: Project
        """
        item = self._client.get_project(project, self.id)
        self._logger.info(f"Project with id: {item.id} successfully retrieved.")
        return Project(item.id, self, item.name, item.description)

    def delete_project(self, project: Reference) -> None:
        """
        Deletes the project specified by the user.

        :param project: The project name or id to delete.

        :return: None
        """
        project_id = self._client.get_project(project, self.id).id
        self._client.delete_project(project_id, self.id)
        self._logger.info(f"Project with id: {project_id} successfully deleted.")

    def list_projects(
        self,
        search: Optional[str] = None,
        page_index: int = 1,
        page_size: int = 20,
    ) -> List[Project]:
        """
        Lists the projects in this workspace.

        :param search: The name to search
        :param page_index: The page index
        :param page_size: The page size

        :return: a list of projects
        """
        response = self._client.list_projects(self.id, search, page_index, page_size)
        return [Project(item.id, self, item.name, item.description) for item in response.list]

    def list_connections(
        self,
        connection_type: Optional[str] = None,
        search: Optional[str] = None,
        page_index: int = 1,
        page_size: int = 20,
    ) -> List[Connection]:
        """
        Lists the connections defined in this workspace.

        :param connection_type: The connection type
        :param search: The name to search
        :param page_index: The page index
        :param page_size: The page size

        :return: A list of connections
        """
        response = self._client.list_connections(self.id, connection_type, search, page_index, page_size)
        return [
            Connection(item.id, item.name, self, item.type, item.parameters, item.description) for item in response.list
        ]

    def get_connection(self, connection: Reference) -> Connection:
        """
        Gets a connection.

        Gets a connection instance with the specified connection name or id in the current workspace.


        :param connection: The connection name or id to get.

        :return: Connection
        """
        item = self._client.get_connection(connection, self.id)
        self._logger.info(f"Connection with id: {item.id} successfully retrieved.")
        return Connection(item.id, item.name, self, item.type, item.parameters, item.description)

    def create_connection(
        self, name: str, type: ConnectionType, parameters: Optional[Dict[str, Any]] = None, description: str = ""
    ) -> Connection:
        """
        Creates a connection.

        Create a connection with the specified connection name, description and
        parameters. The parameters keys and values depends on each connection Type.
        Example of required parameters. {"key_value": data, "file_name": "file.json", "connection_info": data}


        :param name: The name of the connection
        :param type: The type of connection
        :param parameters: The parameters for the connection
        :param description: The description of the connection

        :return: Connection
        """
        data = ConnectionInput(
            name=name,
            parameters=parameters,
            type=type,
            description=description,
        )
        output = self._client.create_connection(data, self.id)
        self._logger.info(f"Connection with id: {output.id} successfully created.")
        return Connection(output.id, output.name, self, output.type, output.parameters, output.description)

    def delete_connection(self, connection: Reference) -> None:
        """
        Deletes the connection.
        :param connection: The connection name or id to delete.
        :return: None
        """
        connection_id = self._client.get_connection(connection, self.id).id
        self._client.delete_connection(connection, self.id)
        self._logger.info(f"Connection with id: {connection_id} successfully deleted.")

    def update_connection(
        self,
        connection: Reference,
        name: Optional[str] = None,
        description: Optional[str] = None,
        parameters: Optional[Dict[str, str]] = None,
    ) -> Connection:
        """
        Updates a connection.
        :param connection: The connection name or id to delete.
        :param name: The new connection name
        :param description: The new connection description
        :param parameters: The new connection parameters
        :return: Connection
        """
        connection_object = self.get_connection(connection)
        connection_input = ConnectionInput(
            name=name if name is not None else connection_object.name,
            type=connection_object.type,
            description=description if description is not None else connection_object.description,
            parameters=parameters,
        )
        connection_update = self._client.update_connection(connection_input, connection, self.id)
        self._logger.info(f"Connection with id: {connection_update.id} successfully updated.")
        return Connection(
            connection_update.id,
            connection_update.name,
            self,
            connection_update.type,
            connection_update.parameters,
            connection_update.description,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Dict

from vectice.api.json.connection_type import ConnectionType

if TYPE_CHECKING:
    from vectice.models import Workspace


class Connection:
    """
    Connection to a repository that contains datasets
    """

    def __init__(
        self,
        id: int,
        name: str,
        workspace: Workspace,
        type: ConnectionType,
        parameters: Dict[str, str],
        description: str = "",
    ):
        """
        :param id: connection identifier
        :param name: name of the connection
        :param workspace: the workspace the connection belong to
        :param type: the type of connection
        :param description: the description of the connection
        """
        self._id = id
        self._name = name
        self._workspace = workspace
        self._type = type
        self._description = description
        self._parameters = parameters

    def __repr__(self):
        return f"Connection(id={self.id}, name={self.name}, type={self.type}, description={self.description}, workspace={self.workspace})"

    @property
    def id(self) -> int:
        """
        Connection identifier.
        :return: int
        """
        return self._id

    @property
    def name(self) -> str:
        """
        Name of the connection.
        """
        return self._name

    @property
    def workspace(self) -> Workspace:
        """
        The workspace this connection belong to.
        :return: Workspace
        """
        return self._workspace

    @property
    def type(self) -> ConnectionType:
        """
        Type of the connection.
        :return: ConnectionType
        """
        return self._type

    @property
    def description(self) -> str:
        """
        Description of the connection.
        :return: str
        """
        return self._description

    @property
    def parameters(self) -> Dict[str, str]:
        """
        Parameters of the connection.
        :return: Dict[str, str]
        """
        return self._parameters.copy()



from .artifact import ArtifactInput, ArtifactOutput, JobArtifactType, ArtifactType
from .artifact_reference import (
    ArtifactReferenceInput,
    RulesDatasetVersionInput,
    RulesModelVersionInput,
    RulesCodeVersionInput,
    ArtifactReferenceOutput,
)
from .artifact_version import ArtifactVersion, VersionStrategy
from .attachment import AttachmentOutput
from .code import CodeInput, CodeOutput
from .code_version import CodeVersionInput, CodeVersionOutput, GitVersionOutput, GitVersionInput
from .connection import ConnectionOutput, ConnectionInput
from .connection_type import ConnectionType
from .data_resource_schema import DataResourceSchema, SchemaColumn, DataType
from .dataset import DatasetInput, DatasetOutput
from .dataset_version import DatasetVersionInput, DatasetVersionOutput
from .files_metadata import FileMetadata, FileMetadataType
from .job import JobOutput, JobInput, JobType
from .metric import MetricInput, MetricOutput
from .model import ModelInput, ModelOutput, ModelType
from .model_version import ModelVersionInput, ModelVersionOutput, ModelVersionStatus
from .page import Page
from .paged_response import PagedResponse
from .project import ProjectOutput, ProjectInput
from .property import PropertyInput, PropertyOutput
from .rule import StopRunOutput, StopRunInput, StartRunInput
from .run import RunInput, RunOutput, RunStatus
from .user_declared_version import UserDeclaredVersion
from .workspace import WorkspaceOutput, WorkspaceInput
from .stage import StageInput, StageOutput, StageOrigin, StageStatus, DocumentationPageType
from .block import BlockInput, BlockOutput, BlockType, BlockState, BlockVariant

__all__ = [
    "ArtifactInput",
    "ArtifactOutput",
    "ArtifactReferenceInput",
    "ArtifactReferenceOutput",
    "ArtifactType",
    "ArtifactVersion",
    "VersionStrategy",
    "ConnectionType",
    "GitVersionInput",
    "GitVersionOutput",
    "JobArtifactType",
    "AttachmentOutput",
    "ConnectionInput",
    "ConnectionOutput",
    "CodeInput",
    "CodeOutput",
    "CodeVersionInput",
    "CodeVersionOutput",
    "DataResourceSchema",
    "DatasetInput",
    "DatasetOutput",
    "DatasetVersionInput",
    "DatasetVersionOutput",
    "DataType",
    "MetricInput",
    "MetricOutput",
    "ModelInput",
    "ModelOutput",
    "ModelType",
    "ModelVersionInput",
    "ModelVersionOutput",
    "ModelVersionStatus",
    "UserDeclaredVersion",
    "JobInput",
    "JobType",
    "JobOutput",
    "JobArtifactType",
    "PagedResponse",
    "ProjectInput",
    "ProjectOutput",
    "PropertyInput",
    "PropertyOutput",
    "RulesCodeVersionInput",
    "RulesDatasetVersionInput",
    "RulesModelVersionInput",
    "RunInput",
    "RunOutput",
    "RunStatus",
    "SchemaColumn",
    "StartRunInput",
    "StopRunOutput",
    "StopRunInput",
    "FileMetadata",
    "FileMetadataType",
    "WorkspaceOutput",
    "WorkspaceInput",
    "Page",
    "StageInput",
    "StageOutput",
    "StageOrigin",
    "StageStatus",
    "DocumentationPageType",
    "BlockInput",
    "BlockOutput",
    "BlockType",
    "BlockState",
    "BlockVariant",
]



from dataclasses import dataclass
from typing import List, Optional

from .artifact_reference import ArtifactReferenceInput, ArtifactReferenceOutput
from .job import JobInput
from .run import RunInput, RunOutput


class StartRunInput:
    def __init__(self, job: Optional[JobInput], run: RunInput, inputs: Optional[List[ArtifactReferenceInput]] = None):
        self._job = job
        self._jobRun = run
        self._jobRun.job = None
        if inputs is not None:
            self._inputArtifacts = [item for item in inputs if item is not None]
        else:
            self._inputArtifacts = []

    @property
    def job(self) -> Optional[JobInput]:
        return self._job

    @property
    def run(self) -> RunInput:
        return self._jobRun

    @property
    def inputs(self) -> List[ArtifactReferenceInput]:
        return self._inputArtifacts


class StopRunInput:
    def __init__(self, run: RunOutput, outputs: Optional[List[ArtifactReferenceInput]] = None):
        self._jobRun = run
        if outputs is not None:
            self._outputArtifacts = [item for item in outputs if item is not None]
        else:
            self._outputArtifacts = []

    @property
    def jobRun(self) -> RunOutput:
        return self._jobRun

    @property
    def outputArtifacts(self) -> Optional[List[ArtifactReferenceInput]]:
        return self._outputArtifacts


class StopRunOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "jobRun" in self:
            self._jobRun: RunOutput = RunOutput(**self["jobRun"])
        if "jobArtifacts" in self:
            self._outputArtifacts: Optional[List[ArtifactReferenceOutput]] = [
                ArtifactReferenceOutput(**item) for item in self["jobArtifacts"] if item is not None
            ]
        else:
            self._outputArtifacts: Optional[List[ArtifactReferenceOutput]] = None

    @property
    def jobRun(self) -> RunOutput:
        return self._jobRun

    @property
    def outputArtifacts(self) -> Optional[List[ArtifactReferenceOutput]]:
        return self._outputArtifacts


@dataclass()
class FillRunInput:
    jobRunId: int
    inputArtifacts: List[ArtifactReferenceInput]
    outputArtifacts: List[ArtifactReferenceInput]


class FillRunOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
